/*
 *  $Id: correct_perspective.c 23069 2021-01-13 15:28:45Z yeti-dn $
 *  Copyright (C) 2021 David Necas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libprocess/gwyprocesstypes.h>
#include <libprocess/arithmetic.h>
#include <libprocess/stats.h>
#include <libprocess/correct.h>
#include <libprocess/elliptic.h>
#include <libgwydgets/gwyradiobuttons.h>
#include <libgwydgets/gwycombobox.h>
#include <libgwydgets/gwystock.h>
#include <libgwymodule/gwymodule-process.h>
#include <app/gwymoduleutils.h>
#include <app/gwyapp.h>
#include "preview.h"

#define PERSPECTIVE_RUN_MODES (GWY_RUN_INTERACTIVE)

typedef enum {
    IMAGE_DATA,
    IMAGE_CORRECTED,
} ImageMode;

typedef struct {
    GwyInterpolationType interp;
    ImageMode image_mode;
    gdouble xy[8];
} PercorArgs;

typedef struct {
    PercorArgs *args;
    GtkWidget *dialog;
    GtkWidget *view;
    GwyVectorLayer *vlayer;
    GwySelection *selection;
    GwyContainer *mydata;
    GSList *image_mode;
    GtkWidget *interp;

    gboolean calculated;
    gulong recalculate_id;
} PercorControls;

static gboolean      module_register                (void);
static void          correct_perspective            (GwyContainer *data,
                                                     GwyRunType run);
static gboolean      percor_dialog                  (PercorArgs *args,
                                                     GwyContainer *data,
                                                     GwyDataField *dfield,
                                                     gint id);
static void          init_coordinates               (GwyDataField *dfield,
                                                     gdouble *xy);
static void          image_mode_changed             (GtkToggleButton *button,
                                                     PercorControls *controls);
static void          selection_changed              (PercorControls *controls);
static void          interp_changed                 (GtkComboBox *combo,
                                                     PercorControls *controls);
static void          invalidate                     (PercorControls *controls);
static gboolean      recalculate                    (gpointer user_data);
static void          do_correction                  (PercorControls *controls);
static GwyDataField* create_corrected_dfield        (GwyDataField *dfield,
                                                     const gdouble *xy,
                                                     GwyInterpolationType interp);
static gboolean      solve_projection_from_rectangle(const gdouble *xy,
                                                     gint xres,
                                                     gint yres,
                                                     gdouble *matrix);
static void          percor_load_args               (GwyContainer *container,
                                                     PercorArgs *args);
static void          percor_save_args               (GwyContainer *container,
                                                     PercorArgs *args);
static void          percor_sanitize_args           (PercorArgs *args);

static const PercorArgs percor_defaults = {
    GWY_INTERPOLATION_LINEAR,
    IMAGE_DATA,
    { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 },
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Corrects or applies perspective distortion of images."),
    "Yeti <yeti@gwyddion.net>",
    "1.0",
    "David Nečas (Yeti)",
    "2021",
};

GWY_MODULE_QUERY2(module_info, correct_perspective)

static gboolean
module_register(void)
{
    gwy_process_func_register("correct_perspective",
                              (GwyProcessFunc)&correct_perspective,
                              N_("/_Distortion/_Perspective..."),
                              NULL,
                              PERSPECTIVE_RUN_MODES,
                              GWY_MENU_FLAG_DATA,
                              N_("Correct perspective distortion"));

    return TRUE;
}

static void
correct_perspective(GwyContainer *data, GwyRunType run)
{
    PercorArgs args;
    GwyDataField *dfield, *mfield, *sfield, *corrected;
    GwySelection *selection;
    GType gtype;
    gint id, newid;
    gchar *s, *t;
    gboolean ok;
    gchar selkey[40];

    g_return_if_fail(run & PERSPECTIVE_RUN_MODES);
    g_return_if_fail(g_type_from_name("GwyLayerProjective"));
    percor_load_args(gwy_app_settings_get(), &args);
    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD, &dfield,
                                     GWY_APP_MASK_FIELD, &mfield,
                                     GWY_APP_SHOW_FIELD, &sfield,
                                     GWY_APP_DATA_FIELD_ID, &id,
                                     0);
    g_return_if_fail(dfield);

    init_coordinates(dfield, args.xy);
    g_snprintf(selkey, sizeof(selkey), "/%d/select/projective", id);
    if (gwy_container_gis_object_by_name(data, selkey, &selection)
        && gwy_selection_get_data(GWY_SELECTION(selection), NULL))
        gwy_selection_get_object(GWY_SELECTION(selection), 0, args.xy);

    ok = percor_dialog(&args, data, dfield, id);
    percor_save_args(gwy_app_settings_get(), &args);

    gtype = g_type_from_name("GwySelectionProjective");
    selection = g_object_new(gtype, "max-objects", 1, NULL);
    gwy_selection_set_object(GWY_SELECTION(selection), 0, args.xy);
    gwy_container_set_object_by_name(data, selkey, selection);
    g_object_unref(selection);

    if (!ok)
        return;

    corrected = create_corrected_dfield(dfield, args.xy, args.interp);
    newid = gwy_app_data_browser_add_data_field(corrected, data, TRUE);
    g_object_unref(corrected);
    s = gwy_app_get_data_field_title(data, id);
    t = g_strconcat(s, " ", _("Corrected"), NULL);
    gwy_container_set_string(data, gwy_app_get_data_title_key_for_id(newid),
                             (const guchar*)t);
    g_free(s);

    if (mfield) {
        corrected = create_corrected_dfield(mfield, args.xy,
                                            GWY_INTERPOLATION_ROUND);
        gwy_container_set_object(data, gwy_app_get_mask_key_for_id(newid),
                                 corrected);
        g_object_unref(corrected);
    }

    if (sfield) {
        corrected = create_corrected_dfield(mfield, args.xy, args.interp);
        gwy_container_set_object(data, gwy_app_get_show_key_for_id(newid),
                                 corrected);
        g_object_unref(corrected);
    }

    gwy_app_sync_data_items(data, data, id, newid, FALSE,
                            GWY_DATA_ITEM_RANGE_TYPE,
                            GWY_DATA_ITEM_RANGE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_MASK_COLOR,
                            0);
    gwy_app_channel_log_add_proc(data, id, newid);
}

static gboolean
percor_dialog(PercorArgs *args,
              GwyContainer *data,
              GwyDataField *dfield,
              gint id)
{
    GtkWidget *hbox, *label, *alignment;
    GtkDialog *dialog;
    GtkTable *table;
    PercorControls controls;
    gint response, row;

    gwy_clear(&controls, 1);
    controls.args = args;

    controls.dialog = gtk_dialog_new_with_buttons(_("Projective Correction"),
                                                  NULL, 0, NULL);
    dialog = GTK_DIALOG(controls.dialog);
    gtk_dialog_add_button(dialog, _("_Reset"), RESPONSE_RESET);
    gtk_dialog_add_button(dialog, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
    gtk_dialog_add_button(dialog, GTK_STOCK_OK, GTK_RESPONSE_OK);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_OK);
    gwy_help_add_to_proc_dialog(GTK_DIALOG(dialog), GWY_HELP_DEFAULT);

    hbox = gtk_hbox_new(FALSE, 2);

    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox,
                       FALSE, FALSE, 4);

    controls.mydata = gwy_container_new();
    gwy_container_set_object_by_name(controls.mydata, "/0/data", dfield);
    gwy_app_sync_data_items(data, controls.mydata, id, 0, FALSE,
                            GWY_DATA_ITEM_RANGE_TYPE,
                            GWY_DATA_ITEM_RANGE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_REAL_SQUARE,
                            0);

    alignment = gtk_alignment_new(0.0, 0.0, 0.0, 0.0);
    gtk_box_pack_start(GTK_BOX(hbox), alignment, FALSE, FALSE, 4);

    controls.view = create_preview(controls.mydata, 0, PREVIEW_SIZE, FALSE);
    controls.selection = create_vector_layer(GWY_DATA_VIEW(controls.view),
                                             0, "Projective", TRUE);
    g_object_ref(controls.selection);
    gwy_selection_set_max_objects(controls.selection, 1);
    controls.vlayer = gwy_data_view_get_top_layer(GWY_DATA_VIEW(controls.view));
    g_object_ref(controls.vlayer);
    g_signal_connect_swapped(controls.selection, "changed",
                             G_CALLBACK(selection_changed), &controls);

    gtk_container_add(GTK_CONTAINER(alignment), controls.view);

    table = GTK_TABLE(gtk_table_new(5, 3, FALSE));
    gtk_table_set_row_spacings(table, 2);
    gtk_table_set_col_spacings(table, 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(table), TRUE, TRUE, 0);
    row = 0;

    label = gtk_label_new(_("Display:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(table, label, 0, 5, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    controls.image_mode
        = gwy_radio_buttons_createl(G_CALLBACK(image_mode_changed), &controls,
                                    args->image_mode,
                                    _("_Data"), IMAGE_DATA,
                                    _("Correc_ted data"), IMAGE_CORRECTED,
                                    NULL);
    row = gwy_radio_buttons_attach_to_table(controls.image_mode,
                                            table, 3, row);

    gtk_table_set_row_spacing(table, row-1, 8);
    label = gtk_label_new_with_mnemonic(_("_Interpolation type:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(table, label, 0, 2, row, row+1, GTK_FILL, 0, 0, 0);

    controls.interp
        = gwy_enum_combo_box_new(gwy_interpolation_type_get_enum(), -1,
                                 G_CALLBACK(interp_changed), &controls,
                                 args->interp, TRUE);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), controls.interp);
    gtk_table_attach(table, controls.interp,
                     2, 5, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    gwy_selection_set_data(controls.selection, 1, args->xy);

    gtk_widget_show_all(controls.dialog);
    do {
        response = gtk_dialog_run(dialog);
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            gtk_widget_destroy(controls.dialog);
            case GTK_RESPONSE_NONE:
            goto finalize;
            break;

            case GTK_RESPONSE_OK:
            break;

            case RESPONSE_RESET:
            init_coordinates(dfield, args->xy);
            gwy_selection_set_data(controls.selection, 1, args->xy);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    gtk_widget_destroy(controls.dialog);

finalize:
    if (controls.recalculate_id)
        g_source_remove(controls.recalculate_id);
    g_object_unref(controls.selection);
    g_object_unref(controls.vlayer);
    g_object_unref(controls.mydata);

    return response == GTK_RESPONSE_OK;
}

static void
init_coordinates(GwyDataField *dfield, gdouble *xy)
{
    gdouble xreal, yreal;
    gint i, j;

    xreal = gwy_data_field_get_xreal(dfield);
    yreal = gwy_data_field_get_yreal(dfield);
    for (i = 0; i < 2; i++) {
        for (j = 0; j < 2; j++) {
            xy[2*(2*i + j) + 0] = 0.25*xreal + 0.5*(i ? 1-j : j)*xreal;
            xy[2*(2*i + j) + 1] = 0.25*yreal + 0.5*i*yreal;
        }
    }
}

static void
image_mode_changed(G_GNUC_UNUSED GtkToggleButton *button,
                   PercorControls *controls)
{
    PercorArgs *args = controls->args;
    GwyDataView *dataview;
    GwyPixmapLayer *layer;
    ImageMode mode;

    mode = gwy_radio_buttons_get_current(controls->image_mode);
    if (mode == args->image_mode)
        return;
    args->image_mode = mode;
    dataview = GWY_DATA_VIEW(controls->view);
    layer = gwy_data_view_get_base_layer(dataview);

    if (args->image_mode == IMAGE_DATA) {
        g_object_set(layer, "data-key", "/0/data", NULL);
        if (!gwy_data_view_get_top_layer(dataview))
            gwy_data_view_set_top_layer(dataview, controls->vlayer);
    }
    else if (args->image_mode == IMAGE_CORRECTED) {
        if (!controls->calculated)
            do_correction(controls);
        g_object_set(layer, "data-key", "/1/data", NULL);
        gwy_data_view_set_top_layer(dataview, NULL);
    }

    gwy_set_data_preview_size(dataview, PREVIEW_SIZE);
}

static void
selection_changed(PercorControls *controls)
{
    gwy_selection_get_object(controls->selection, 0, controls->args->xy);
    invalidate(controls);
}

static void
interp_changed(GtkComboBox *combo,
               PercorControls *controls)
{
    controls->args->interp = gwy_enum_combo_box_get_active(combo);
    invalidate(controls);
}

static void
invalidate(PercorControls *controls)
{
    controls->calculated = FALSE;
    if (controls->args->image_mode != IMAGE_CORRECTED)
        return;

    if (controls->recalculate_id)
        return;

    controls->recalculate_id = g_idle_add(recalculate, controls);
}

static gboolean
recalculate(gpointer user_data)
{
    PercorControls *controls = (PercorControls*)user_data;
    do_correction(controls);
    controls->recalculate_id = 0;
    return FALSE;
}

static void
do_correction(PercorControls *controls)
{
    PercorArgs *args = controls->args;
    GwyDataField *dfield, *corrected;
    gdouble xy[8];

    dfield = GWY_DATA_FIELD(gwy_container_get_object_by_name(controls->mydata,
                                                             "/0/data"));
    gwy_selection_get_object(controls->selection, 0, xy);
    corrected = create_corrected_dfield(dfield, xy, args->interp);
    gwy_container_set_object_by_name(controls->mydata, "/1/data", corrected);
    g_object_unref(corrected);

    controls->calculated = TRUE;
}

static void
project(gdouble x, gdouble y, gdouble *px, gdouble *py, gpointer user_data)
{
    const gdouble *matrix = (const gdouble*)user_data;
    const gdouble *mx = matrix, *my = matrix + 3, *m1 = matrix + 6;
    gdouble d;

    d = m1[0]*x + m1[1]*y + m1[2];
    *px = (mx[0]*x + mx[1]*y + mx[2])/d;
    *py = (my[0]*x + my[1]*y + my[2])/d;
}

static void
estimate_reasonable_dimensions(const gdouble *xy, gdouble *lx, gdouble *ly)
{
    gdouble lx1, lx2, ly1, ly2;

    lx1 = hypot(xy[2] - xy[0], xy[3] - xy[1]);
    lx2 = hypot(xy[6] - xy[4], xy[7] - xy[5]);
    ly1 = hypot(xy[4] - xy[2], xy[5] - xy[3]);
    ly2 = hypot(xy[0] - xy[6], xy[1] - xy[5]);

    *lx = hypot(lx1, lx2)/G_SQRT2;
    *ly = hypot(ly1, ly2)/G_SQRT2;
}

static GwyDataField*
create_corrected_dfield(GwyDataField *dfield,
                        const gdouble *xy,
                        GwyInterpolationType interp)
{
    GwyDataField *corrected;
    gdouble matrix[9], xypix[8];
    gdouble newxreal, newyreal, newxpix, newypix;
    gint i, newxres, newyres;

    for (i = 0; i < 4; i++) {
        xypix[2*i + 0] = gwy_data_field_rtoj(dfield, xy[2*i + 0]);
        xypix[2*i + 1] = gwy_data_field_rtoi(dfield, xy[2*i + 1]);
    }
    estimate_reasonable_dimensions(xy, &newxreal, &newyreal);
    estimate_reasonable_dimensions(xypix, &newxpix, &newypix);
    newxres = MAX(GWY_ROUND(newxpix + 1.0), 2);
    newyres = MAX(GWY_ROUND(newypix + 1.0), 2);
    corrected = gwy_data_field_new(newxres, newyres,
                                   newxreal*newxres/newxpix,
                                   newyreal*newyres/newypix,
                                   FALSE);
    gwy_data_field_copy_units(dfield, corrected);

    solve_projection_from_rectangle(xypix, newxres, newyres, matrix);
    gwy_data_field_distort(dfield, corrected, project, matrix, interp,
                           GWY_EXTERIOR_MIRROR_EXTEND, 0.0);

    return corrected;
}

static gboolean
solve_projection(const gdouble *xyfrom,
                 const gdouble *xyto,
                 gdouble *matrix)
{
    gdouble a[64], rhs[8];
    guint i;

    gwy_clear(a, 64);
    for (i = 0; i < 4; i++) {
        gdouble xf = xyfrom[2*i + 0], yf = xyfrom[2*i + 1];
        gdouble xt = xyto[2*i + 0], yt = xyto[2*i + 1];
        gdouble *axrow = a + 16*i, *ayrow = axrow + 8, *r = rhs + 2*i;

        axrow[0] = ayrow[3] = xf;
        axrow[1] = ayrow[4] = yf;
        axrow[2] = ayrow[5] = 1.0;
        axrow[6] = -xf*xt;
        axrow[7] = -yf*xt;
        ayrow[6] = -xf*yt;
        ayrow[7] = -yf*yt;
        r[0] = xt;
        r[1] = yt;
    }

    if (!gwy_math_lin_solve_rewrite(8, a, rhs, matrix))
        return FALSE;

    matrix[8] = 1.0;
    return TRUE;
}

static gboolean
solve_projection_from_rectangle(const gdouble *xy,
                                gint xres, gint yres,
                                gdouble *matrix)
{
    gdouble rectangle[8];

    rectangle[0] = rectangle[1] = rectangle[3] = rectangle[6] = 0.5;
    rectangle[2] = rectangle[4] = xres - 0.5;
    rectangle[5] = rectangle[7] = yres - 0.5;

    return solve_projection(rectangle, xy, matrix);
}

static const gchar interp_key[] = "/module/correct_perspective/interpolation";

static void
percor_sanitize_args(PercorArgs *args)
{
    args->interp = gwy_enum_sanitize_value(args->interp,
                                           GWY_TYPE_INTERPOLATION_TYPE);
}

static void
percor_load_args(GwyContainer *container,
                 PercorArgs *args)
{
    *args = percor_defaults;

    gwy_container_gis_enum_by_name(container, interp_key, &args->interp);

    percor_sanitize_args(args);
}

static void
percor_save_args(GwyContainer *container,
                 PercorArgs *args)
{
    gwy_container_set_enum_by_name(container, interp_key, args->interp);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
