/*
 *  $Id: pat_synth.c 23117 2021-01-29 08:18:25Z yeti-dn $
 *  Copyright (C) 2010-2021 David Necas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libgwyddion/gwythreads.h>
#include <libgwyddion/gwyrandgenset.h>
#include <libprocess/correct.h>
#include <libprocess/stats.h>
#include <libprocess/filters.h>
#include <libgwydgets/gwystock.h>
#include <libgwymodule/gwymodule-process.h>
#include <app/gwyapp.h>
#include "dimensions.h"
#include "preview.h"

#define PAT_SYNTH_RUN_MODES (GWY_RUN_IMMEDIATE | GWY_RUN_INTERACTIVE)

#define DECLARE_PATTERN(name) \
    static gpointer create_gui_##name(PatSynthControls *controls); \
    static void dimensions_changed_##name(PatSynthControls *controls); \
    static void make_pattern_##name(const PatSynthArgs *args, \
                                    const GwyDimensionArgs *dimsargs, \
                                    GwyRandGenSet *rngset, \
                                    GwyDataField *dfield); \
    static gpointer load_args_##name(GwyContainer *settings); \
    static void save_args_##name(gpointer pargs, GwyContainer *settings); \
    static void reset_##name(gpointer pcontrols)

#define PATTERN_FUNCS(name) \
    &create_gui_##name, &dimensions_changed_##name, &reset_##name, \
    &make_pattern_##name, \
    &load_args_##name, &save_args_##name

enum {
    PAGE_DIMENSIONS = 0,
    PAGE_GENERATOR  = 1,
    PAGE_PLACEMENT  = 2,
    PAGE_NPAGES
};

/* Each pattern has its own set of parameters but many are common so they get
 * the same symbolic name in PatSynthRng for simpliciy. */
typedef enum {
    /* Deformation (displacement map) is universal. */
    RNG_DISPLAC_X   = 0,
    RNG_DISPLAC_Y   = 1,
    RNG_HEIGHT      = 2,
    RNG_TOP_X       = 3,
    RNG_SIZE_X      = 3,  /* Alias. */
    RNG_TOP_Y       = 4,
    RNG_SLOPE       = 5,
    RNG_OFFSET_X    = 6,
    RNG_OFFSET_Y    = 7,
    RNG_ROUNDNESS   = 8,
    RNG_ORIENTATION = 8,  /* Alias */
    RNG_NRNGS
} PatSynthRng;

typedef enum {
    PAT_SYNTH_STAIRCASE = 0,
    PAT_SYNTH_DBLSTAIR  = 1,
    PAT_SYNTH_GRATING   = 2,
    PAT_SYNTH_AMPHITH   = 3,
    PAT_SYNTH_RINGS     = 4,
    PAT_SYNTH_STAR      = 5,
    PAT_SYNTH_RHOLES    = 6,
    PAT_SYNTH_PILLARS   = 7,
    PAT_SYNTH_NTYPES
} PatSynthType;

typedef enum {
    PILLAR_SHAPE_CIRCLE  = 0,
    PILLAR_SHAPE_SQUARE  = 1,
    PILLAR_SHAPE_HEXAGON = 2,
    PILLAR_SHAPE_NSHAPES,
} PillarShapeType;

typedef struct _PatSynthArgs PatSynthArgs;
typedef struct _PatSynthControls PatSynthControls;

typedef gpointer (*CreateGUIFunc)        (PatSynthControls *controls);
typedef void     (*DestroyGUIFunc)       (gpointer pcontrols);
typedef void     (*DimensionsChangedFunc)(PatSynthControls *controls);
typedef void     (*ResetFunc)            (gpointer pcontrols);
typedef void     (*MakePatternFunc)      (const PatSynthArgs *args,
                                          const GwyDimensionArgs *dimsargs,
                                          GwyRandGenSet *rngset,
                                          GwyDataField *dfield);
typedef gpointer (*LoadArgsFunc)         (GwyContainer *settings);
typedef void     (*SaveArgsFunc)         (gpointer pargs,
                                          GwyContainer *settings);

/* This scheme makes the object type list easily reordeable in the GUI without
 * changing the ids.  */
typedef struct {
    PatSynthType type;
    const gchar *name;
    CreateGUIFunc create_gui;
    DimensionsChangedFunc dims_changed;
    ResetFunc reset;
    MakePatternFunc run;
    LoadArgsFunc load_args;
    SaveArgsFunc save_args;
} PatSynthPattern;

struct _PatSynthArgs {
    gint active_page;
    gint seed;
    gboolean randomize;
    gboolean update;
    PatSynthType type;
    gpointer pattern_args;
};

struct _PatSynthControls {
    PatSynthArgs *args;
    GwyDimensions *dims;
    const PatSynthPattern *pattern;
    GwyRandGenSet *rngset;
    /* They have different types, known only to the pattern. */
    gpointer pattern_controls;
    GtkWidget *dialog;
    GtkWidget *view;
    GtkWidget *update;
    GtkWidget *update_now;
    GtkObject *seed;
    GtkWidget *randomize;
    GtkTable *table;
    GtkTable *table_parameters;
    GtkTable *table_placement;
    GtkWidget *type;
    GwyContainer *mydata;
    GwyDataField *surface;
    gdouble pxsize;
    gdouble zscale;
    gboolean in_init;
    gulong sid;
};

static gboolean      module_register      (void);
static void          pat_synth            (GwyContainer *data,
                                           GwyRunType run);
static void          run_noninteractive   (PatSynthArgs *args,
                                           const GwyDimensionArgs *dimsargs,
                                           GwyRandGenSet *rngset,
                                           GwyContainer *data,
                                           GwyDataField *dfield,
                                           gint oldid,
                                           GQuark quark);
static gboolean      pat_synth_dialog     (PatSynthArgs *args,
                                           GwyDimensionArgs *dimsargs,
                                           GwyRandGenSet *rngset,
                                           GwyContainer *data,
                                           GwyDataField *dfield,
                                           gint id);
static GtkWidget*    pattern_selector_new (PatSynthControls *controls);
static void          update_controls      (PatSynthControls *controls,
                                           PatSynthArgs *args);
static void          page_switched        (PatSynthControls *controls,
                                           GtkNotebookPage *page,
                                           gint pagenum);
static void          update_values        (PatSynthControls *controls);
static void          pattern_type_selected(GtkComboBox *combo,
                                           PatSynthControls *controls);
static void          pat_synth_invalidate (PatSynthControls *controls);
static gboolean      preview_gsource      (gpointer user_data);
static void          preview              (PatSynthControls *controls);
static void          pat_synth_do         (const PatSynthArgs *args,
                                           const GwyDimensionArgs *dimsargs,
                                           GwyRandGenSet *rngset,
                                           GwyDataField *dfield);
static GwyDataField* make_displacement_map(guint xres,
                                           guint yres,
                                           gdouble sigma,
                                           gdouble tau,
                                           GwyRandGenSet *rngset,
                                           guint rngid);
static void          pat_synth_load_args  (GwyContainer *container,
                                           PatSynthArgs *args,
                                           GwyDimensionArgs *dimsargs);
static void          pat_synth_save_args  (GwyContainer *container,
                                           const PatSynthArgs *args,
                                           const GwyDimensionArgs *dimsargs);

#define GWY_SYNTH_CONTROLS PatSynthControls
#define GWY_SYNTH_INVALIDATE(controls) \
    pat_synth_invalidate(controls)

#include "synth.h"

DECLARE_PATTERN(staircase);
DECLARE_PATTERN(dblstair);
DECLARE_PATTERN(grating);
DECLARE_PATTERN(rholes);
DECLARE_PATTERN(pillars);
DECLARE_PATTERN(amphith);
DECLARE_PATTERN(rings);
DECLARE_PATTERN(star);

static const gchar prefix[] = "/module/pat_synth";

static const PatSynthArgs pat_synth_defaults = {
    PAGE_DIMENSIONS,
    42, TRUE, TRUE,
    PAT_SYNTH_STAIRCASE,
    NULL,
};

static const GwyDimensionArgs dims_defaults = GWY_DIMENSION_ARGS_INIT;

static const PatSynthPattern patterns[] = {
    { PAT_SYNTH_STAIRCASE, N_("Staircase"),           PATTERN_FUNCS(staircase), },
    { PAT_SYNTH_GRATING,   N_("Grating"),             PATTERN_FUNCS(grating),   },
    { PAT_SYNTH_AMPHITH,   N_("Amphitheater"),        PATTERN_FUNCS(amphith),   },
    { PAT_SYNTH_RINGS,     N_("Concentric rings"),    PATTERN_FUNCS(rings),     },
    { PAT_SYNTH_STAR,      N_("Siemens star"),        PATTERN_FUNCS(star),      },
    { PAT_SYNTH_RHOLES,    N_("Holes (rectangular)"), PATTERN_FUNCS(rholes),    },
    { PAT_SYNTH_PILLARS,   N_("Pillars"),             PATTERN_FUNCS(pillars),   },
    { PAT_SYNTH_DBLSTAIR,  N_("Double staircase"),    PATTERN_FUNCS(dblstair),  },
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Generates surfaces representing simple patterns "
       "(staircase, amphitheater, grating, holes and pillars, ...)."),
    "Yeti <yeti@gwyddion.net>",
    "2.0",
    "David Nečas (Yeti)",
    "2010",
};

GWY_MODULE_QUERY2(module_info, pat_synth)

static gboolean
module_register(void)
{
    gwy_process_func_register("pat_synth",
                              (GwyProcessFunc)&pat_synth,
                              N_("/S_ynthetic/_Pattern..."),
                              GWY_STOCK_SYNTHETIC_PATTERN,
                              PAT_SYNTH_RUN_MODES,
                              0,
                              N_("Generate patterned surface"));

    return TRUE;
}

static void
pat_synth(GwyContainer *data, GwyRunType run)
{
    PatSynthArgs args;
    GwyDimensionArgs dimsargs;
    GwyRandGenSet *rngset;
    GwyDataField *dfield;
    GQuark quark;
    gboolean ok = TRUE;
    gint id;

    g_return_if_fail(run & PAT_SYNTH_RUN_MODES);
    pat_synth_load_args(gwy_app_settings_get(), &args, &dimsargs);
    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD, &dfield,
                                     GWY_APP_DATA_FIELD_ID, &id,
                                     GWY_APP_DATA_FIELD_KEY, &quark,
                                     0);

    rngset = gwy_rand_gen_set_new(RNG_NRNGS);
    if (run == GWY_RUN_INTERACTIVE) {
        ok = pat_synth_dialog(&args, &dimsargs, rngset, data, dfield, id);
        pat_synth_save_args(gwy_app_settings_get(), &args, &dimsargs);
    }
    if (ok)
        run_noninteractive(&args, &dimsargs, rngset, data, dfield, id, quark);

    gwy_rand_gen_set_free(rngset);
    gwy_dimensions_free_args(&dimsargs);
}

static const PatSynthPattern*
get_pattern(guint type)
{
    guint i;

    for (i = 0; i < G_N_ELEMENTS(patterns); i++) {
        if (patterns[i].type == type)
            return patterns + i;
    }
    g_warning("Unknown pattern %u\n", type);

    return patterns + 0;
}

static void
run_noninteractive(PatSynthArgs *args,
                   const GwyDimensionArgs *dimsargs,
                   GwyRandGenSet *rngset,
                   GwyContainer *data,
                   GwyDataField *dfield,
                   gint oldid,
                   GQuark quark)
{
    const PatSynthPattern *pattern;
    GwySIUnit *siunit;
    gboolean replace = dimsargs->replace && dfield;
    gboolean add = dimsargs->add && dfield;
    gint newid;

    if (args->randomize)
        args->seed = g_random_int() & 0x7fffffff;

    if (replace) {
        /* Always take a reference so that we can always unref. */
        g_object_ref(dfield);

        gwy_app_undo_qcheckpointv(data, 1, &quark);
        if (!add)
            gwy_data_field_clear(dfield);

        gwy_app_channel_log_add_proc(data, oldid, oldid);
    }
    else {
        if (add)
            dfield = gwy_data_field_duplicate(dfield);
        else {
            gdouble mag = pow10(dimsargs->xypow10) * dimsargs->measure;
            dfield = gwy_data_field_new(dimsargs->xres, dimsargs->yres,
                                        mag*dimsargs->xres, mag*dimsargs->yres,
                                        TRUE);

            siunit = gwy_data_field_get_si_unit_xy(dfield);
            gwy_si_unit_set_from_string(siunit, dimsargs->xyunits);

            siunit = gwy_data_field_get_si_unit_z(dfield);
            gwy_si_unit_set_from_string(siunit, dimsargs->zunits);
        }
    }

    pattern = get_pattern(args->type);
    args->pattern_args = pattern->load_args(gwy_app_settings_get());
    pat_synth_do(args, dimsargs, rngset, dfield);
    g_free(args->pattern_args);
    args->pattern_args = NULL;

    if (!replace) {
        if (data) {
            newid = gwy_app_data_browser_add_data_field(dfield, data, TRUE);
            if (oldid != -1)
                gwy_app_sync_data_items(data, data, oldid, newid, FALSE,
                                        GWY_DATA_ITEM_GRADIENT,
                                        0);
        }
        else {
            newid = 0;
            data = gwy_container_new();
            gwy_container_set_object(data, gwy_app_get_data_key_for_id(newid),
                                     dfield);
            gwy_app_data_browser_add(data);
            gwy_app_data_browser_reset_visibility(data,
                                                  GWY_VISIBILITY_RESET_SHOW_ALL);
            g_object_unref(data);
        }

        gwy_app_set_data_field_title(data, newid, _("Generated"));
        gwy_app_channel_log_add_proc(data, add ? oldid : -1, newid);
    }
    g_object_unref(dfield);
}

static gboolean
pat_synth_dialog(PatSynthArgs *args,
                 GwyDimensionArgs *dimsargs,
                 GwyRandGenSet *rngset,
                 GwyContainer *data,
                 GwyDataField *dfield_template,
                 gint id)
{
    GtkWidget *dialog, *table, *vbox, *hbox, *notebook;
    PatSynthControls controls;
    GwyDataField *dfield;
    gboolean finished;
    gint response;
    gint row;

    gwy_clear(&controls, 1);
    controls.in_init = TRUE;
    controls.args = args;
    controls.rngset = rngset;
    controls.pxsize = 1.0;
    dialog = gtk_dialog_new_with_buttons(_("Pattern"),
                                         NULL, 0,
                                         _("_Reset"), RESPONSE_RESET,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    gwy_help_add_to_proc_dialog(GTK_DIALOG(dialog), GWY_HELP_DEFAULT);
    controls.dialog = dialog;

    hbox = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox,
                       FALSE, FALSE, 4);

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 4);

    controls.mydata = gwy_container_new();
    dfield = gwy_data_field_new(PREVIEW_SIZE, PREVIEW_SIZE,
                                dimsargs->measure*PREVIEW_SIZE,
                                dimsargs->measure*PREVIEW_SIZE,
                                TRUE);
    gwy_container_set_object_by_name(controls.mydata, "/0/data", dfield);
    if (dfield_template) {
        gwy_app_sync_data_items(data, controls.mydata, id, 0, FALSE,
                                GWY_DATA_ITEM_PALETTE,
                                0);
        controls.surface = gwy_synth_surface_for_preview(dfield_template,
                                                         PREVIEW_SIZE);
        controls.zscale = 3.0*gwy_data_field_get_rms(dfield_template);
    }
    controls.view = create_preview(controls.mydata, 0, PREVIEW_SIZE, FALSE);
    gtk_box_pack_start(GTK_BOX(vbox), controls.view, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX(vbox),
                       gwy_synth_instant_updates_new(&controls,
                                                     &controls.update_now,
                                                     &controls.update,
                                                     &args->update),
                       FALSE, FALSE, 0);
    g_signal_connect_swapped(controls.update_now, "clicked",
                             G_CALLBACK(preview), &controls);

    gtk_box_pack_start(GTK_BOX(vbox),
                       gwy_synth_random_seed_new(&controls,
                                                 &controls.seed, &args->seed),
                       FALSE, FALSE, 0);

    controls.randomize = gwy_synth_randomize_new(&args->randomize);
    gtk_box_pack_start(GTK_BOX(vbox), controls.randomize, FALSE, FALSE, 0);

    notebook = gtk_notebook_new();
    gtk_box_pack_start(GTK_BOX(hbox), notebook, TRUE, TRUE, 4);
    g_signal_connect_swapped(notebook, "switch-page",
                             G_CALLBACK(page_switched), &controls);

    controls.dims = gwy_dimensions_new(dimsargs, dfield_template);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             gwy_dimensions_get_widget(controls.dims),
                             gtk_label_new(_("Dimensions")));
    if (controls.dims->add)
        g_signal_connect_swapped(controls.dims->add, "toggled",
                                 G_CALLBACK(pat_synth_invalidate), &controls);

    table = gtk_table_new(1, 3, FALSE);
    controls.table_parameters = GTK_TABLE(table);
    gtk_table_set_row_spacings(controls.table_parameters, 2);
    gtk_table_set_col_spacings(controls.table_parameters, 6);
    gtk_table_set_row_spacing(controls.table_parameters, 0, 0);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), table,
                             gtk_label_new(_("Generator")));
    row = 0;

    controls.type = pattern_selector_new(&controls);
    gwy_table_attach_adjbar(table, row, _("_Pattern:"), NULL,
                            GTK_OBJECT(controls.type),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;
    g_object_set_data(G_OBJECT(table), "base-rows", GINT_TO_POINTER(row));

    table = gtk_table_new(1, 4, FALSE);
    controls.table_placement = GTK_TABLE(table);
    gtk_table_set_row_spacings(controls.table_placement, 2);
    gtk_table_set_col_spacings(controls.table_placement, 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), table,
                             gtk_label_new(_("Placement")));
    g_object_set_data(G_OBJECT(table), "base-rows", GINT_TO_POINTER(1));

    controls.table = GTK_TABLE(controls.table_parameters);
    pattern_type_selected(GTK_COMBO_BOX(controls.type), &controls);

    gtk_widget_show_all(dialog);
    controls.in_init = FALSE;
    /* Must be done when widgets are shown, see GtkNotebook docs */
    gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), args->active_page);
    update_values(&controls);
    pat_synth_invalidate(&controls);

    finished = FALSE;
    while (!finished) {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            case GTK_RESPONSE_OK:
            gtk_widget_destroy(dialog);
            case GTK_RESPONSE_NONE:
            finished = TRUE;
            break;

            case RESPONSE_RESET:
            args->seed = pat_synth_defaults.seed;
            args->randomize = pat_synth_defaults.randomize;
            /* Don't reset type either.  It sort of defeats resetting the
             * pattern-specific options. */
            controls.in_init = TRUE;
            controls.pattern->reset(controls.pattern_controls);
            update_controls(&controls, args);
            controls.in_init = FALSE;
            if (args->update)
                preview(&controls);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    }

    pattern_type_selected(NULL, &controls);
    if (controls.sid) {
        g_source_remove(controls.sid);
        controls.sid = 0;
    }
    g_object_unref(controls.mydata);
    GWY_OBJECT_UNREF(controls.surface);
    gwy_dimensions_free(controls.dims);

    return response == GTK_RESPONSE_OK;
}

static GtkWidget*
pattern_selector_new(PatSynthControls *controls)
{
    GtkWidget *combo;
    GwyEnum *model;
    guint n, i;

    n = G_N_ELEMENTS(patterns);
    model = g_new(GwyEnum, n);
    for (i = 0; i < n; i++) {
        model[i].value = patterns[i].type;
        model[i].name = patterns[i].name;
    }

    combo = gwy_enum_combo_box_new(model, n,
                                   G_CALLBACK(pattern_type_selected), controls,
                                   controls->args->type, TRUE);
    g_object_weak_ref(G_OBJECT(combo), (GWeakNotify)g_free, model);

    return combo;
}

static void
update_controls(PatSynthControls *controls,
                PatSynthArgs *args)
{
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(controls->update),
                                 args->update);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->seed), args->seed);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(controls->randomize),
                                 args->randomize);
    gwy_enum_combo_box_set_active(GTK_COMBO_BOX(controls->type), args->type);
}

static void
page_switched(PatSynthControls *controls,
              G_GNUC_UNUSED GtkNotebookPage *page,
              gint pagenum)
{
    if (controls->in_init)
        return;

    controls->args->active_page = pagenum;
    if (pagenum == PAGE_GENERATOR)
        update_values(controls);
}

static void
update_values(PatSynthControls *controls)
{
    GwyDimensions *dims = controls->dims;

    controls->pxsize = dims->args->measure * pow10(dims->args->xypow10);
    controls->pattern->dims_changed(controls);
}

static void
pattern_type_selected(GtkComboBox *combo,
                      PatSynthControls *controls)
{
    const PatSynthPattern *pattern;
    PatSynthArgs *args = controls->args;
    GtkTable *table;
    guint baserows;

    if (controls->pattern) {
        pattern = controls->pattern;
        pattern->save_args(args->pattern_args, gwy_app_settings_get());
        controls->pattern = NULL;
        g_free(controls->pattern_controls);
        controls->pattern_controls = NULL;
        g_free(args->pattern_args);
        args->pattern_args = NULL;
    }

    /* Just tear-down */
    if (!combo)
        return;

    args->type = gwy_enum_combo_box_get_active(combo);
    pattern = controls->pattern = get_pattern(args->type);

    table = controls->table_parameters;
    baserows = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(table), "base-rows"));
    gwy_synth_shrink_table(table, baserows);

    table = controls->table_placement;
    baserows = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(table), "base-rows"));
    gwy_synth_shrink_table(table, baserows);

    args->pattern_args = pattern->load_args(gwy_app_settings_get());
    controls->pattern_controls = pattern->create_gui(controls);
    gtk_widget_show_all(GTK_WIDGET(controls->table_parameters));
    gtk_widget_show_all(GTK_WIDGET(controls->table_placement));

    update_values(controls);
    pat_synth_invalidate(controls);
}

static void
pat_synth_invalidate(PatSynthControls *controls)
{
    /* create preview if instant updates are on */
    if (controls->args->update && !controls->in_init && !controls->sid) {
        controls->sid = g_idle_add_full(G_PRIORITY_LOW, preview_gsource,
                                        controls, NULL);
    }
}

static gboolean
preview_gsource(gpointer user_data)
{
    PatSynthControls *controls = (PatSynthControls*)user_data;
    controls->sid = 0;

    preview(controls);

    return FALSE;
}

static void
preview(PatSynthControls *controls)
{
    PatSynthArgs *args = controls->args;
    GwyDataField *dfield;

    dfield = GWY_DATA_FIELD(gwy_container_get_object_by_name(controls->mydata,
                                                             "/0/data"));
    if (controls->dims->args->add && controls->surface)
        gwy_data_field_copy(controls->surface, dfield, FALSE);
    else
        gwy_data_field_clear(dfield);

    pat_synth_do(args, controls->dims->args, controls->rngset, dfield);
}

static void
pat_synth_do(const PatSynthArgs *args,
             const GwyDimensionArgs *dimsargs,
             GwyRandGenSet *rngset,
             GwyDataField *dfield)
{
    const PatSynthPattern *pattern = get_pattern(args->type);

    gwy_rand_gen_set_init(rngset, args->seed);
    pattern->run(args, dimsargs, rngset, dfield);
    gwy_data_field_data_changed(dfield);
}

/***************************************************************************
 *
 * Common helpers
 *
 ***************************************************************************/

/* Iterating through square in a spiral fashion from the origin to preserve the
 * centre conrer if it's randomly generated.  Field @k holds the current index
 * in the two-dimensional array.  */
typedef struct {
    gint n;
    gint i, j, k;
    gint istep, jstep;
    gint s, segmentend, ntotalstep;
} GrowingIter;

static inline void
growing_iter_init(GrowingIter *giter, guint n)
{
    giter->n = n;
    giter->j = giter->i = 0;
    giter->istep = 0;
    giter->jstep = -1;
    giter->ntotalstep = n*n;
    giter->segmentend = MIN(1, n*n);
    giter->s = 0;
    giter->k = (n/2 - giter->i)*n + (giter->j + n/2);
}

static inline gboolean
growing_iter_next(GrowingIter *giter)
{
    giter->i += giter->istep;
    giter->j += giter->jstep;
    giter->k = (giter->n/2 - giter->i)*giter->n + (giter->j + giter->n/2);
    giter->s++;
    if (giter->s == giter->segmentend) {
        if (giter->s == giter->ntotalstep)
            return FALSE;

        if (giter->i == giter->j + 1) {
            giter->istep = 1;
            giter->jstep = 0;
            giter->segmentend = 1 - 2*giter->i;
        }
        else if (giter->i == giter->j) {
            giter->istep = -1;
            giter->jstep = 0;
            giter->segmentend = 2*giter->i;
        }
        else if (giter->j > 0) {
            giter->istep = 0;
            giter->jstep = -1;
            giter->segmentend = 2*giter->j + 1;
        }
        else {
            giter->istep = 0;
            giter->jstep = 1;
            giter->segmentend = 2*giter->i;
        }
        giter->segmentend += giter->s;
        giter->segmentend = MIN(giter->segmentend, giter->ntotalstep);
    }
    return TRUE;
}

static guint
bisect_lower(const gdouble *a, guint n, gdouble x)
{
    guint lo = 0, hi = n-1;

    if (G_UNLIKELY(x < a[lo]))
        return 0;
    if (G_UNLIKELY(x >= a[hi]))
        return n-1;

    while (hi - lo > 1) {
        guint mid = (hi + lo)/2;

        if (x < a[mid])
            hi = mid;
        else
            lo = mid;
    }

    return lo;
}

static inline gdouble
superellipse(gdouble x, gdouble y, gdouble invpower)
{
    gdouble m, M;

    x = fabs(x);
    y = fabs(y);
    m = fmin(x, y);
    M = fmax(x, y);
    /* This is t = (x^n + y^n)^{1/n}.
     * However, we rewrite is as x(1 + (y/x)^n)^{1/n} for x ≥ y, making the
     * parenthesis converging quickly to 1 for large n, making all the powers
     * numerically safe. */
    return M*pow(1.0 + pow(m/M, 2.0/invpower), invpower/2.0);
}

static gdouble
parabolic_transform(gdouble x, gdouble alpha)
{
    if (alpha > 0.0)
        return (1.0 - alpha)*x + alpha*x*x;
    if (alpha < 0.0) {
        gdouble a1 = 1.0 + alpha;
        return 2.0*x/(sqrt(a1*a1 - 4.0*alpha*x) + a1);
    }
    return x;
}

/* Random number within the range [-1/2, 1/2] going from bell shape for small
 * s to uniform for large s. */
static inline gdouble
random_constrained_shift(GwyRandGenSet *rngset, guint rngid, gdouble s)
{
    gdouble r, ss;

    r = gwy_rand_gen_set_double(rngset, rngid);
    ss = s*4.6;
    if (ss < 1.0)
        return ss/G_PI * asin(2.0*r - 1.0);
    return 0.5/asin(1.0/ss) * asin((2.0*r - 1.0)/ss);
}

/* Fill a data field with uncorrelated random numbers in a growing fashion to
 * preserve the character of the noise even if the dimensions change */
/* XXX: Duplicate with lat_synth */
static void
fill_displacement_map(GwyDataField *dfield,
                      gdouble q,
                      GwyRandGenSet *rngset, guint rngid)
{
    GrowingIter giter;
    guint xres, yres;
    gdouble *data;

    xres = gwy_data_field_get_xres(dfield);
    yres = gwy_data_field_get_yres(dfield);
    g_return_if_fail(xres == yres);
    data = gwy_data_field_get_data(dfield);
    growing_iter_init(&giter, xres);

    do {
        data[giter.k] = q*(gwy_rand_gen_set_double(rngset, rngid) - 0.5);
    } while (growing_iter_next(&giter));
}

static GwyDataField*
make_displacement_map(guint xres, guint yres,
                      gdouble sigma, gdouble tau,
                      GwyRandGenSet *rngset, guint rngid)
{
    GwyDataField *grid, *dfield;
    guint gn, n;
    gdouble q, r;

    n = MAX(xres, yres);
    q = 2.0*sigma*tau;
    if (!q)
        return gwy_data_field_new(xres, yres, 1.0, 1.0, TRUE);

    if (tau <= 1.0) {
        grid = gwy_data_field_new(n, n, 1.0, 1.0, FALSE);
        fill_displacement_map(grid, q, rngset, rngid);
        gwy_data_field_filter_gaussian(grid, tau);
        if (xres == yres)
            return grid;

        dfield = gwy_data_field_area_extract(grid, 0, 0, xres, yres);
        g_object_unref(grid);
        return dfield;
    }

    gn = GWY_ROUND(1.0/tau*n);
    gn = MAX(gn, 2);
    r = (gdouble)gn/n;
    grid = gwy_data_field_new(gn, gn, 1.0, 1.0, FALSE);
    fill_displacement_map(grid, q*r, rngset, rngid);
    gwy_data_field_filter_gaussian(grid, r*tau);
    dfield = gwy_data_field_new_resampled(grid, n, n, GWY_INTERPOLATION_KEY);
    g_object_unref(grid);

    if (xres != yres) {
        grid = dfield;
        dfield = gwy_data_field_area_extract(grid, 0, 0, xres, yres);
        g_object_unref(grid);
    }

    return dfield;
}

/* Usually tmap is the same data field as displacement_x. */
static void
displacement_to_t_linear(GwyDataField *tmap,
                         GwyDataField *displacement_x,
                         GwyDataField *displacement_y,
                         gdouble angle,
                         gdouble period)
{
    gdouble *xdata = gwy_data_field_get_data(displacement_x);
    gdouble *ydata = gwy_data_field_get_data(displacement_y);
    gdouble *tdata = gwy_data_field_get_data(tmap);
    guint xres = gwy_data_field_get_xres(tmap);
    guint yres = gwy_data_field_get_yres(tmap);
    gdouble c = cos(angle), s = sin(angle);
    gdouble toff = 0.5*(s*(yres - 1) - c*(xres - 1));
    guint i;

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(xdata,ydata,tdata,toff,c,s,period,xres,yres)
#endif
    for (i = 0; i < yres; i++) {
        gdouble *xrow = xdata + i*xres;
        gdouble *yrow = ydata + i*xres;
        gdouble *trow = tdata + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble t = toff;

            t += (j + xrow[j])*c - (i + yrow[j])*s;
            t /= period;
            trow[j] = t;
        }
    }
}

/* Usually tmap is the same data field as displacement_x. */
static void
displacement_to_t_superellipse(GwyDataField *tmap,
                               GwyDataField *displacement_x,
                               GwyDataField *displacement_y,
                               gdouble angle,
                               gdouble xcentre,
                               gdouble ycentre,
                               gdouble invpower,
                               gdouble radius)
{
    gdouble *xdata = gwy_data_field_get_data(displacement_x);
    gdouble *ydata = gwy_data_field_get_data(displacement_y);
    gdouble *tdata = gwy_data_field_get_data(tmap);
    guint xres = gwy_data_field_get_xres(tmap);
    guint yres = gwy_data_field_get_yres(tmap);
    gdouble c = cos(angle), s = sin(angle);
    guint i;

    /* Account for percieved step direction being along the diagonals, not
     * along the axes, and so the steps seen as narrower. */
    if (invpower > 1.0)
        radius *= pow(2.0, invpower/2.0)/G_SQRT2;

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(xdata,ydata,tdata,xcentre,ycentre,invpower,radius,c,s,xres,yres)
#endif
    for (i = 0; i < yres; i++) {
        gdouble *xrow = xdata + i*xres;
        gdouble *yrow = ydata + i*xres;
        gdouble *trow = tdata + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble x, y, t;

            x = j + xrow[j] - 0.5*(xres - 1) - xres*xcentre;
            y = i + yrow[j] - 0.5*(yres - 1) - yres*ycentre;
            t = x*c - y*s;
            y = x*s + y*c;
            x = t;

            if (invpower < 0.000001)
                t = fmax(fabs(x), fabs(y));
            else if (invpower > 1.999999)
                t = fabs(x) + fabs(y);
            else if (invpower > 0.999999 && invpower < 1.000001)
                t = sqrt(x*x + y*y);
            else
                t = superellipse(x, y, invpower);
            trow[j] = t/radius;
        }
    }
}

/* Usually umap and vmap are the same data field as displacement_x and y. */
static void
displacement_to_uv_linear(GwyDataField *umap,
                          GwyDataField *vmap,
                          GwyDataField *displacement_x,
                          GwyDataField *displacement_y,
                          gdouble angle,
                          gdouble periodu,
                          gdouble periodv)
{
    gdouble *xdata = gwy_data_field_get_data(displacement_x);
    gdouble *ydata = gwy_data_field_get_data(displacement_y);
    gdouble *udata = gwy_data_field_get_data(umap);
    gdouble *vdata = gwy_data_field_get_data(vmap);
    guint xres = gwy_data_field_get_xres(umap);
    guint yres = gwy_data_field_get_yres(umap);
    gdouble c = cos(angle), s = sin(angle);
    gdouble uoff = 0.5*(s*(yres - 1) - c*(xres - 1));
    gdouble voff = -0.5*(c*(xres - 1) + s*(xres - 1));
    guint i;

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(xdata,ydata,udata,vdata,uoff,voff,c,s,periodu,periodv,xres,yres)
#endif
    for (i = 0; i < yres; i++) {
        gdouble *xrow = xdata + i*xres;
        gdouble *yrow = ydata + i*xres;
        gdouble *urow = udata + i*xres;
        gdouble *vrow = vdata + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble x = j + xrow[j], y = i + yrow[j];

            urow[j] = (x*c - y*s + uoff)/periodu;
            vrow[j] = (x*s + y*c + voff)/periodv;
        }
    }
}

static guint
find_t_range(GwyDataField *tmap, gboolean positive)
{
    gdouble tmin, tmax, tt;
    guint n;

    gwy_data_field_get_min_max(tmap, &tmin, &tmax);
    if (positive) {
        /* Cover the range [0, tmax]. */
        g_warn_if_fail(tmin >= 0.0);
        n = (GWY_ROUND(tmax + 3.5) | 1);
    }
    else {
        /* Cover a symmetrical range (presumably tmin ≈ -tmax). */
        tt = fmax(tmax, -tmin);
        n = 2*GWY_ROUND(tt + 3.5) + 1;
    }
    gwy_debug("tmin = %g, tmax = %g, n = %u", tmin, tmax, n);

    return n;
}

static gdouble*
make_values_1d(guint n,
               gdouble mean, gdouble noise,
               GwyRandGenSet *rngset, guint rngid)
{
    gdouble *values;
    guint i, centre = n/2;
    gdouble r;

    g_return_val_if_fail(n & 1, NULL);
    values = g_new(gdouble, n);

    values[centre] = mean*gwy_rand_gen_set_multiplier(rngset, rngid, noise);
    for (i = 1; i <= n/2; i++) {
        r = gwy_rand_gen_set_multiplier(rngset, rngid, noise);
        values[centre + i] = mean*r;
        r = gwy_rand_gen_set_multiplier(rngset, rngid, noise);
        values[centre - i] = mean*r;
    }

    return values;
}

static gdouble*
make_values_2d(guint n,
               gdouble mean, gdouble noise,
               GwyRandGenSet *rngset, guint rngid)
{
    GrowingIter giter;
    gdouble *values;
    gdouble r;

    g_return_val_if_fail(n & 1, NULL);
    values = g_new(gdouble, n*n);
    growing_iter_init(&giter, n);

    do {
        r = gwy_rand_gen_set_multiplier(rngset, rngid, noise);
        values[giter.k] = mean*r;
    } while (growing_iter_next(&giter));

    return values;
}

static gdouble*
make_values_2d_gaussian(guint n,
                        gdouble mean, gdouble noise,
                        GwyRandGenSet *rngset, guint rngid)
{
    GrowingIter giter;
    gdouble *values;
    gdouble r;

    g_return_val_if_fail(n & 1, NULL);
    values = g_new(gdouble, n*n);
    growing_iter_init(&giter, n);

    do {
        r = gwy_rand_gen_set_gaussian(rngset, rngid, noise);
        values[giter.k] = mean + r;
    } while (growing_iter_next(&giter));

    return values;
}

static gdouble*
distribute_left_to_left_and_right(gdouble *left, guint n, gdouble asymm)
{
    gdouble *right;
    guint i;

    right = g_new(gdouble, n);
    for (i = 0; i < n; i++) {
        right[i] = 0.5*(1.0 + asymm)*left[i];
        left[i] = 0.5*(1.0 - asymm)*left[i];
    }

    return right;
}

static void
transform_to_scaled_grating(gdouble *abscissae, gdouble *widths,
                            gdouble *leftslopes, gdouble *rightslopes,
                            guint n,
                            gboolean zero_based)
{
    gdouble *newabscissae;
    gdouble w, c, a0 = sizeof("GCC is broken");
    guint i;

    newabscissae = g_new(gdouble, n);
    if (zero_based) {
        a0 = abscissae[0];
        abscissae[0] = -abscissae[-1];
    }
    for (i = 0; i < n; i++) {
        if (!i) {
            w = abscissae[i+1] - abscissae[i];
            c = abscissae[i];
        }
        else if (i == n-1) {
            w = abscissae[i] - abscissae[i-1];
            c = abscissae[i];
        }
        else {
            w = 0.5*(abscissae[i+1] - abscissae[i-1]);
            c = 0.5*abscissae[i] + 0.25*(abscissae[i+1] + abscissae[i-1]);
        }
        widths[i] *= w;
        leftslopes[i] *= w;
        rightslopes[i] *= w;
        newabscissae[i] = c + 0.5*(leftslopes[i] - rightslopes[i]);
    }
    if (zero_based)
        newabscissae[0] = a0;
    gwy_assign(abscissae, newabscissae, n);
    g_free(newabscissae);
}

static void
transform_to_sine_cosine(gdouble *angles_sines, gdouble *cosines,
                         guint n)
{
    guint i;

    for (i = 0; i < n; i++) {
        cosines[i] = cos(angles_sines[i]);
        angles_sines[i] = sin(angles_sines[i]);
    }
}

static gdouble*
make_positions_1d_linear(guint n,
                         gdouble noise,
                         GwyRandGenSet *rngset, guint rngid)
{
    gdouble *abscissae;
    guint i, centre = n/2;
    gdouble r;

    g_return_val_if_fail(n & 1, NULL);
    abscissae = g_new(gdouble, n);

    /* Fill the positions from centre for stability. */
    r = random_constrained_shift(rngset, rngid, noise);
    abscissae[centre] = r;
    for (i = 1; i <= n/2; i++) {
        r = random_constrained_shift(rngset, rngid, noise);
        abscissae[centre + i] = r + i;
        r = random_constrained_shift(rngset, rngid, noise);
        abscissae[centre - i] = r - i;
    }

    return abscissae;
}

static gdouble*
make_positions_1d_radial(guint n,
                         gdouble noise, gdouble scale, gdouble parabolicity,
                         GwyRandGenSet *rngset, guint rngid)
{
    gdouble *radii;
    gdouble r, a;
    guint i;

    g_return_val_if_fail(n & 1, NULL);
    radii = g_new(gdouble, n);

    radii[0] = -100.0;
    for (i = 1; i < n; i++) {
        r = random_constrained_shift(rngset, rngid, noise);
        a = parabolic_transform(i/scale, -parabolicity)*scale;
        radii[i] = r + a;
    }

    return radii;
}

static gdouble*
make_positions_2d_linear(guint n,
                         gdouble noise, gboolean is_y,
                         GwyRandGenSet *rngset, guint rngid)
{
    gdouble *abscissae;
    gdouble r;
    GrowingIter giter;

    g_return_val_if_fail(n & 1, NULL);
    abscissae = g_new(gdouble, n*n);
    growing_iter_init(&giter, n);

    do {
        r = random_constrained_shift(rngset, rngid, noise);
        abscissae[giter.k] = r + (is_y ? -giter.i : giter.j);
    } while (growing_iter_next(&giter));

    return abscissae;
}

/* abscissae[] has n elements, but heights[] have n+1. */
static gdouble*
make_heights_staircase(const gdouble *abscissae, guint n,
                       gdouble h, gdouble noise,
                       gboolean keep_slope, gboolean zero_based,
                       GwyRandGenSet *rngset, guint rngid)
{
    gdouble *heights;
    guint i, centre = n/2;
    gdouble r;

    g_return_val_if_fail(n & 1, NULL);
    heights = g_new(gdouble, n+1);

    /* Generate n steps. */
    if (keep_slope) {
        heights[0] = heights[n] = h;
        for (i = 1; i < n; i++)
            heights[i] = 0.5*h*(abscissae[i+1] - abscissae[i-1]);
    }
    else {
        for (i = 1; i <= n; i++)
            heights[i] = h;
    }

    heights[centre+1] *= gwy_rand_gen_set_multiplier(rngset, rngid, noise);
    for (i = 1; i <= n/2; i++) {
        r = gwy_rand_gen_set_multiplier(rngset, rngid, noise);
        heights[centre+1 + i] *= r;
        r = gwy_rand_gen_set_multiplier(rngset, rngid, noise);
        heights[centre+1 - i] *= r;
    }

    /* Convert them to n+1 absolute heights. */
    heights[0] = 0.0;
    for (i = 1; i <= n; i++)
        heights[i] += heights[i-1];

    if (!zero_based) {
        h = 0.5*(heights[centre] + heights[centre+1]);
        for (i = 0; i <= n; i++)
            heights[i] -= h;
    }

    return heights;
}

static inline gdouble
step_func(gdouble x, gdouble w)
{
    if (G_LIKELY(w != 0.0))
        return fmin(fmax(x/w + 0.5, 0.0), 1.0);
    return x > 0.0 ? 1.0 : (x < 0.0 ? -1.0 : 0.0);
}

static inline gdouble
ridge_func(gdouble x, gdouble w, gdouble sleft, gdouble sright)
{
    if (x < -0.5*w) {
        x += 0.5*w;
        return x <= -sleft ? 0.0 : 1.0 + x/sleft;
    }
    if (x > 0.5*w) {
        x -= 0.5*w;
        return x >= sright ? 0.0 : 1.0 - x/sright;
    }
    return 1.0;
}

/* It is actually inverted hole to keep the sign convention. */
static inline gdouble
hole_func(gdouble x, gdouble y, gdouble ax, gdouble ay, gdouble r, gdouble s)
{
    gdouble d;

    x = fabs(x);
    y = fabs(y);

    if (ax < ay) {
        GWY_SWAP(gdouble, ax, ay);
        GWY_SWAP(gdouble, x, y);
    }
    r = fmin(r, ay);

    if (x <= ax - r && x-y <= ax - ay)
        d = ay - y;
    else if (y <= ay - r && x-y >= ax - ay)
        d = ax - x;
    else {
        x -= ax - r;
        y -= ay - r;
        d = r - sqrt(x*x + y*y);
    }

    if (s == 0.0)
        return d > 0.0;

    return fmax(fmin(d/s, 1.0), 0.0);
}

static void
render_staircase(GwyDataField *image, GwyDataField *tmap,
                 const gdouble *abscissae,
                 const gdouble *heights,
                 const gdouble *slopes,
                 guint n)
{
    const gdouble *tdata;
    gdouble *data;
    guint i, xres, yres;

    xres = gwy_data_field_get_xres(image);
    yres = gwy_data_field_get_yres(image);
    tdata = gwy_data_field_get_data_const(tmap);
    data = gwy_data_field_get_data(image);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,tdata,abscissae,heights,slopes,xres,yres,n)
#endif
    for (i = 0; i < yres; i++) {
        const gdouble *trow = tdata + i*xres;
        gdouble *row = data + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble z, t, s, x, w;
            guint k, m, mfrom, mto;

            t = trow[j];
            k = bisect_lower(abscissae, n, t);
            mfrom = (k < 1 ? 0: k-1);
            mto = (k >= n-2 ? n-1 : k+2);
            z = heights[mfrom];
            for (m = mfrom; m <= mto; m++) {
                s = heights[m+1] - heights[m];
                x = abscissae[m];
                w = slopes[m];
                z += s*step_func(t - x, w);
            }
            row[j] += z;
        }
    }
}

static void
render_double_staircase(GwyDataField *image,
                        GwyDataField *umap,
                        GwyDataField *vmap,
                        const gdouble *abscissaeu,
                        const gdouble *abscissaev,
                        const gdouble *heights,
                        guint nu,
                        guint nv)
{
    const gdouble *udata, *vdata;
    gdouble *data;
    guint i, n, xres, yres;

    xres = gwy_data_field_get_xres(image);
    yres = gwy_data_field_get_yres(image);
    udata = gwy_data_field_get_data_const(umap);
    vdata = gwy_data_field_get_data_const(vmap);
    data = gwy_data_field_get_data(image);
    n = ((nu + nv) | 1);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,udata,vdata,abscissaeu,abscissaev,heights,xres,yres,nu,nv,n)
#endif
    for (i = 0; i < yres; i++) {
        const gdouble *urow = udata + i*xres;
        const gdouble *vrow = vdata + i*xres;
        gdouble *row = data + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble z, u, v;
            guint ku, kv;

            u = urow[j];
            v = vrow[j];
            ku = bisect_lower(abscissaeu, nu, u);
            kv = bisect_lower(abscissaev, nv, v);
            z = heights[MIN(ku + kv, n-1)];
            row[j] += z;
        }
    }
}

static void
render_grating(GwyDataField *image, GwyDataField *tmap,
               const gdouble *abscissae,
               const gdouble *widths,
               const gdouble *heights,
               const gdouble *leftslopes,
               const gdouble *rightslopes,
               gint sign,
               guint n)
{
    const gdouble *tdata;
    gdouble *data;
    guint i, xres, yres;

    xres = gwy_data_field_get_xres(image);
    yres = gwy_data_field_get_yres(image);
    tdata = gwy_data_field_get_data_const(tmap);
    data = gwy_data_field_get_data(image);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,tdata,abscissae,heights,widths,leftslopes,rightslopes,xres,yres,sign,n)
#endif
    for (i = 0; i < yres; i++) {
        const gdouble *trow = tdata + i*xres;
        gdouble *row = data + i*xres;
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble z, t, h, sl, sr, x, w;
            guint k, m, mfrom, mto;

            t = trow[j];
            k = bisect_lower(abscissae, n, t);
            mfrom = (k < 1 ? 0: k-1);
            mto = (k >= n-2 ? n-1 : k+2);
            z = 0;
            for (m = mfrom; m <= mto; m++) {
                x = abscissae[m];
                h = heights[m];
                w = widths[m];
                sl = leftslopes[m];
                sr = rightslopes[m];
                z = fmax(z, h*ridge_func(t - x, w, sl, sr));
            }
            row[j] += sign*z;
        }
    }
}

static void
render_holes(GwyDataField *image,
             GwyDataField *umap,
             GwyDataField *vmap,
             const gdouble *abscissaeu,
             const gdouble *abscissaev,
             const gdouble *xsizes,
             const gdouble *ysizes,
             const gdouble *slopes,
             const gdouble *roundnesses,
             const gdouble *heights,
             guint n,
             gdouble aratio)
{
    const gdouble *udata, *vdata;
    gdouble *data;
    guint i, xres, yres;
    gdouble ax, ay;

    if (aratio <= 1.0) {
        ax = 1.0;
        ay = aratio;
    }
    else {
        ax = 1.0/aratio;
        ay = 1.0;
    }

    xres = gwy_data_field_get_xres(image);
    yres = gwy_data_field_get_yres(image);
    udata = gwy_data_field_get_data_const(umap);
    vdata = gwy_data_field_get_data_const(vmap);
    data = gwy_data_field_get_data(image);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,udata,vdata,abscissaeu,abscissaev,xsizes,ysizes,slopes,roundnesses,heights,xres,yres,n,ax,ay)
#endif
    for (i = 0; i < yres; i++) {
        const gdouble *urow = udata + i*xres;
        const gdouble *vrow = vdata + i*xres;
        gdouble *row = data + i*xres;
        gdouble amin = fmin(ax, ay);
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble z = 0.0, u, v, x, y, h, lx, ly, s, r;
            guint ku, kv, mufrom, muto, mvfrom, mvto, mu, mv, kk;

            u = urow[j];
            v = vrow[j];
            ku = (guint)(n/2 + floor(u));
            kv = (guint)(n/2 + floor(v));
            ku = MIN(ku, n-1);
            kv = MIN(kv, n-1);
            mufrom = (ku < 1 ? 0: ku-1);
            muto = (ku >= n-2 ? n-1 : ku+2);
            mvfrom = (kv < 1 ? 0: kv-1);
            mvto = (kv >= n-2 ? n-1 : kv+2);
            for (mv = mvfrom; mv <= mvto; mv++) {
                for (mu = mufrom; mu <= muto; mu++) {
                    kk = mv*n + mu;
                    h = heights[kk];
                    s = slopes[kk] * amin;
                    r = roundnesses[kk] * amin;
                    y = v - abscissaev[kk];
                    y = (2.0*y - 1) * ay;
                    ly = ysizes[kk] * ay;
                    x = u - abscissaeu[kk];
                    x = (2.0*x - 1) * ax;
                    lx = xsizes[kk] * ax;
                    z = fmax(z, h*hole_func(x, y, lx, ly, r, s));
                }
            }
            row[j] -= z;
        }
    }
}

static void
render_pillars(GwyDataField *image,
               GwyDataField *umap,
               GwyDataField *vmap,
               PillarShapeType shape,
               const gdouble *abscissaeu,
               const gdouble *abscissaev,
               const gdouble *sizes,
               const gdouble *slopes,
               const gdouble *sines,
               const gdouble *cosines,
               const gdouble *heights,
               guint n,
               gdouble aratio)
{
    const gdouble *udata, *vdata;
    gdouble *data;
    guint i, xres, yres;
    gdouble ax, ay;

    if (aratio <= 1.0) {
        ax = 1.0;
        ay = aratio;
    }
    else {
        ax = 1.0/aratio;
        ay = 1.0;
    }

    xres = gwy_data_field_get_xres(image);
    yres = gwy_data_field_get_yres(image);
    udata = gwy_data_field_get_data_const(umap);
    vdata = gwy_data_field_get_data_const(vmap);
    data = gwy_data_field_get_data(image);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,udata,vdata,shape,abscissaeu,abscissaev,sizes,slopes,sines,cosines,heights,xres,yres,n,ax,ay)
#endif
    for (i = 0; i < yres; i++) {
        const gdouble *urow = udata + i*xres;
        const gdouble *vrow = vdata + i*xres;
        gdouble *row = data + i*xres;
        gdouble amin = fmin(ax, ay);
        guint j;

        for (j = 0; j < xres; j++) {
            gdouble z = 0.0, u, v, h, w, s, t, x, y, ca, sa;
            guint ku, kv, mufrom, muto, mvfrom, mvto, mu, mv, kk;

            u = urow[j];
            v = vrow[j];
            ku = (guint)(n/2 + floor(u));
            kv = (guint)(n/2 + floor(v));
            ku = MIN(ku, n-1);
            kv = MIN(kv, n-1);
            mufrom = (ku < 1 ? 0: ku-1);
            muto = (ku >= n-2 ? n-1 : ku+2);
            mvfrom = (kv < 1 ? 0: kv-1);
            mvto = (kv >= n-2 ? n-1 : kv+2);
            for (mv = mvfrom; mv <= mvto; mv++) {
                for (mu = mufrom; mu <= muto; mu++) {
                    kk = mv*n + mu;
                    h = heights[kk];
                    w = sizes[kk] * amin;
                    s = slopes[kk] * amin;
                    y = v - abscissaev[kk];
                    y = (2.0*y - 1) * ay;
                    x = u - abscissaeu[kk];
                    x = (2.0*x - 1) * ax;

                    if (shape == PILLAR_SHAPE_CIRCLE)
                        t = sqrt(x*x + y*y);
                    else {
                        ca = cosines[kk];
                        sa = sines[kk];
                        if (shape == PILLAR_SHAPE_SQUARE)
                            t = fmax(fabs(ca*x - sa*y), fabs(sa*x + ca*y));
                        else {
                            t = ca*x - sa*y;
                            y = sa*x + ca*y;
                            x = t;
                            t = fmax(fabs(x),
                                     fmax(0.5*fabs(x + GWY_SQRT3*y),
                                          0.5*fabs(x - GWY_SQRT3*y)));
                        }
                    }

                    if (s == 0.0)
                        t = (t <= w);
                    else
                        t = fmax(fmin(1.0 + (w - t)/s, 1.0), 0.0);
                    z = fmax(z, h*t);
                }
            }
            row[j] += z;
        }
    }
}

static gint
attach_superellipse_invpower(PatSynthControls *controls,
                             gint row,
                             GtkObject **adj,
                             gdouble *target)
{
    *adj = gtk_adjustment_new(*target, 0.0, 2.0, 0.001, 0.1, 0);
    g_object_set_data(G_OBJECT(*adj), "target", target);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("Super_ellipse parameter:"), NULL,
                            GTK_OBJECT(*adj), GWY_HSCALE_LINEAR);
    g_signal_connect_swapped(*adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);

    return row;
}

static gint
attach_parabolicity(PatSynthControls *controls,
                    gint row,
                    GtkObject **adj,
                    gdouble *target)
{
    *adj = gtk_adjustment_new(*target, -1.0, 1.0, 0.0001, 0.1, 0);
    g_object_set_data(G_OBJECT(*adj), "target", target);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("_Parabolicity:"), NULL,
                            GTK_OBJECT(*adj), GWY_HSCALE_LINEAR);
    g_signal_connect_swapped(*adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);

    return row;
}

static gint
attach_asymmetry(PatSynthControls *controls,
                 gint row,
                 GtkObject **adj,
                 gdouble *target)
{
    *adj = gtk_adjustment_new(*target, -1.0, 1.0, 0.001, 0.1, 0);
    g_object_set_data(G_OBJECT(*adj), "target", target);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("_Asymmetry:"), NULL,
                            GTK_OBJECT(*adj), GWY_HSCALE_LINEAR);
    g_signal_connect_swapped(*adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);

    return row;
}

static gint
attach_centre_offset(PatSynthControls *controls,
                     gint row,
                     GtkObject **adj,
                     gdouble *target,
                     gboolean is_y)
{
    *adj = gtk_adjustment_new(*target, -2.0, 2.0, 0.01, 0.1, 0);
    g_object_set_data(G_OBJECT(*adj), "target", target);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            is_y ? _("_Y center:") : _("_X center:"), NULL,
                            GTK_OBJECT(*adj), GWY_HSCALE_SQRT);
    g_signal_connect_swapped(*adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);

    return row;
}

/* The following terrible mess implements a lateral quantity which is selected
 * as a fraction of [the minimum of two] other quantities. */
static void
update_lateral_frac_of_min(GtkAdjustment *adj)
{
    PatSynthControls *controls = g_object_get_data(G_OBJECT(adj), "controls");
    GtkWidget *label = g_object_get_data(G_OBJECT(adj), "value-label");
    GtkAdjustment *master1 = g_object_get_data(G_OBJECT(adj), "master1");
    GtkAdjustment *master2 = g_object_get_data(G_OBJECT(adj), "master2");
    gdouble val = gtk_adjustment_get_value(adj);
    gdouble masterval1 = gtk_adjustment_get_value(master1);
    gdouble masterval2 = gtk_adjustment_get_value(master2 ? master2 : master1);
    gdouble scale = fmin(masterval1, masterval2) * controls->pxsize;

    gwy_synth_update_value_label(GTK_LABEL(label), controls->dims->sxyvf,
                                 val*scale);
}

G_GNUC_UNUSED
static gint
attach_lateral_frac(PatSynthControls *controls,
                    gint row,
                    GtkObject *adj,
                    gdouble *target,
                    const gchar *name,
                    GwyHScaleStyle hscale_style,
                    GtkObject *master,
                    GtkWidget **pspin,
                    GtkWidget **punits)
{
    GtkWidget *spin, *value_label;

    g_object_set_data(G_OBJECT(adj), "target", target);
    g_object_set_data(G_OBJECT(adj), "controls", controls);
    g_object_set_data(G_OBJECT(adj), "master1", master);

    spin = gwy_table_attach_adjbar(GTK_WIDGET(controls->table),
                                   row, name, NULL, adj, hscale_style);
    if (pspin)
        *pspin = spin;
    row++;

    value_label = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(value_label), 1.0, 0.5);
    gtk_table_attach(controls->table, value_label,
                     1, 2, row, row+1, GTK_FILL, 0, 0, 0);
    g_object_set_data(G_OBJECT(adj), "value-label", value_label);

    *punits = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(*punits), 0.0, 0.5);
    gtk_table_attach(controls->table, *punits,
                     2, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    g_signal_connect_swapped(adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(update_lateral_frac_of_min), NULL);
    g_signal_connect_swapped(master, "value-changed",
                             G_CALLBACK(update_lateral_frac_of_min), adj);

    return row;
}

static gint
attach_lateral_frac_of_min(PatSynthControls *controls,
                           gint row,
                           GtkObject *adj,
                           gdouble *target,
                           const gchar *name,
                           GwyHScaleStyle hscale_style,
                           GtkObject *master1,
                           GtkObject *master2,
                           GtkWidget **pspin,
                           GtkWidget **punits)
{
    GtkWidget *spin, *value_label;

    g_object_set_data(G_OBJECT(adj), "target", target);
    g_object_set_data(G_OBJECT(adj), "controls", controls);
    g_object_set_data(G_OBJECT(adj), "master1", master1);
    g_object_set_data(G_OBJECT(adj), "master2", master2);

    spin = gwy_table_attach_adjbar(GTK_WIDGET(controls->table),
                                   row, name, NULL, adj, hscale_style);
    if (pspin)
        *pspin = spin;
    row++;

    value_label = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(value_label), 1.0, 0.5);
    gtk_table_attach(controls->table, value_label,
                     1, 2, row, row+1, GTK_FILL, 0, 0, 0);
    g_object_set_data(G_OBJECT(adj), "value-label", value_label);

    *punits = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(*punits), 0.0, 0.5);
    gtk_table_attach(controls->table, *punits,
                     2, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    g_signal_connect_swapped(adj, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(update_lateral_frac_of_min), NULL);
    g_signal_connect_swapped(master1, "value-changed",
                             G_CALLBACK(update_lateral_frac_of_min), adj);
    g_signal_connect_swapped(master2, "value-changed",
                             G_CALLBACK(update_lateral_frac_of_min), adj);

    return row;
}

/***************************************************************************
 *
 * Staircase
 *
 ***************************************************************************/

typedef struct {
    gdouble period;
    gdouble position_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble height;
    gdouble height_noise;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
    gboolean keep_slope;
} PatSynthArgsStaircase;

typedef struct {
    PatSynthArgsStaircase *args;
    GtkObject *period;
    GtkWidget *period_units;
    GtkObject *position_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkWidget *keep_slope;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsStaircase;

static const PatSynthArgsStaircase pat_synth_defaults_staircase = {
    50.0, 0.0,
    0.05, 0.0,
    1.0, 0.0,
    0.0,
    0.0, 10.0,
    FALSE,
};

static gpointer
create_gui_staircase(PatSynthControls *controls)
{
    PatSynthControlsStaircase *pcontrols;
    PatSynthArgsStaircase *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 4+4+4);
    pcontrols = g_new0(PatSynthControlsStaircase, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    row = gwy_synth_attach_heading(controls, row, _("Terrace"), NULL);
    pcontrols->period = gtk_adjustment_new(pargs->period,
                                           1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->period, &pargs->period,
                                   _("_Terrace width:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->period_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->position_noise,
                                             &pargs->position_noise);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->slope, &pargs->slope,
                              _("_Slope width:"), GWY_HSCALE_SQRT,
                              pcontrols->period, NULL,
                              &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);
    row = gwy_synth_attach_checkbox(controls, row,
                                    &pcontrols->keep_slope, &pargs->keep_slope,
                                    _("Scales _with width"));

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_staircase(PatSynthControls *controls)
{
    PatSynthControlsStaircase *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->period_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->period));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_staircase(const PatSynthArgs *args,
                       const GwyDimensionArgs *dimsargs,
                       GwyRandGenSet *rngset,
                       GwyDataField *dfield)
{
    const PatSynthArgsStaircase *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    gdouble *abscissa, *height, *slope;
    GwyDataField *displx, *disply, *tmap;
    guint n;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    tmap = displx;  /* Alias for clarity. */
    displacement_to_t_linear(tmap, displx, disply, pargs->angle, pargs->period);
    n = find_t_range(tmap, FALSE);
    abscissa = make_positions_1d_linear(n, pargs->position_noise,
                                        rngset, RNG_OFFSET_X);
    height = make_heights_staircase(abscissa, n,
                                    hmean, pargs->height_noise,
                                    pargs->keep_slope, FALSE,
                                    rngset, RNG_HEIGHT);
    slope = make_values_1d(n, pargs->slope, pargs->slope_noise,
                           rngset, RNG_SLOPE);
    render_staircase(dfield, tmap, abscissa, height, slope, n);

    g_free(slope);
    g_free(height);
    g_free(abscissa);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_staircase(GwyContainer *settings)
{
    PatSynthArgsStaircase *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_staircase,
                     sizeof(PatSynthArgsStaircase));
    key = g_string_new(prefix);
    g_string_append(key, "/staircase/");
    gwy_synth_load_arg_double(settings, key, "period", 1.0, 2000.0,
                              &pargs->period);
    gwy_synth_load_arg_double(settings, key, "position_noise", 0.0, 1.0,
                              &pargs->position_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    gwy_synth_load_arg_boolean(settings, key, "keep_slope",
                               &pargs->keep_slope);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_staircase(gpointer p,
                    GwyContainer *settings)
{
    PatSynthArgsStaircase *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/staircase/");
    gwy_synth_save_arg_double(settings, key, "period", pargs->period);
    gwy_synth_save_arg_double(settings, key, "position_noise",
                              pargs->position_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    gwy_synth_save_arg_boolean(settings, key, "keep_slope", pargs->keep_slope);
    g_string_free(key, TRUE);
}

static void
reset_staircase(gpointer p)
{
    PatSynthControlsStaircase *pcontrols = p;
    PatSynthArgsStaircase *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_staircase;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->period), pargs->period);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->position_noise),
                             pargs->position_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pcontrols->keep_slope),
                                 pargs->keep_slope);
}

/***************************************************************************
 *
 * Double staircase
 *
 ***************************************************************************/

typedef struct {
    gdouble xperiod;
    gdouble xposition_noise;
    gdouble yperiod;
    gdouble yposition_noise;
    gdouble height;
    gdouble height_noise;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
} PatSynthArgsDblstair;

typedef struct {
    PatSynthArgsDblstair *args;
    GtkObject *xperiod;
    GtkWidget *xperiod_units;
    GtkObject *xposition_noise;
    GtkObject *yperiod;
    GtkWidget *yperiod_units;
    GtkObject *yposition_noise;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsDblstair;

static const PatSynthArgsDblstair pat_synth_defaults_dblstair = {
    50.0, 0.0,
    50.0, 0.0,
    1.0, 0.0,
    0.0,
    0.0, 10.0,
};

static gpointer
create_gui_dblstair(PatSynthControls *controls)
{
    PatSynthControlsDblstair *pcontrols;
    PatSynthArgsDblstair *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 7+4+4);
    pcontrols = g_new0(PatSynthControlsDblstair, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    row = gwy_synth_attach_heading(controls, row, _("Terrace"), NULL);
    pcontrols->xperiod = gtk_adjustment_new(pargs->xperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->xperiod, &pargs->xperiod,
                                   _("_Terrace X width:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->xperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->xposition_noise,
                                             &pargs->xposition_noise);
    pcontrols->yperiod = gtk_adjustment_new(pargs->yperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->yperiod, &pargs->yperiod,
                                   _("_Terrace Y width:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->yperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->yposition_noise,
                                             &pargs->yposition_noise);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_dblstair(PatSynthControls *controls)
{
    PatSynthControlsDblstair *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->xperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->yperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->xperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->yperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_dblstair(const PatSynthArgs *args,
                      const GwyDimensionArgs *dimsargs,
                      GwyRandGenSet *rngset,
                      GwyDataField *dfield)
{
    const PatSynthArgsDblstair *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    gdouble *abscissau, *abscissav, *height;
    GwyDataField *displx, *disply, *umap, *vmap;
    guint nu, nv;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    umap = displx;  /* Alias for clarity. */
    vmap = disply;  /* Alias for clarity. */
    displacement_to_uv_linear(umap, vmap, displx, disply, pargs->angle,
                              pargs->xperiod, pargs->yperiod);
    nu = find_t_range(umap, FALSE);
    nv = find_t_range(vmap, FALSE);
    abscissau = make_positions_1d_linear(nu, pargs->xposition_noise,
                                         rngset, RNG_OFFSET_X);
    abscissav = make_positions_1d_linear(nv, pargs->yposition_noise,
                                         rngset, RNG_OFFSET_Y);
    height = make_heights_staircase(NULL, (nu + nv) | 1,
                                    hmean, pargs->height_noise,
                                    FALSE, FALSE,
                                    rngset, RNG_HEIGHT);
    render_double_staircase(dfield, umap, vmap, abscissau, abscissav,
                            height, nu, nv);

    g_free(height);
    g_free(abscissau);
    g_free(abscissav);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_dblstair(GwyContainer *settings)
{
    PatSynthArgsDblstair *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_dblstair,
                     sizeof(PatSynthArgsDblstair));
    key = g_string_new(prefix);
    g_string_append(key, "/dblstair/");
    gwy_synth_load_arg_double(settings, key, "xperiod", 1.0, 2000.0,
                              &pargs->xperiod);
    gwy_synth_load_arg_double(settings, key, "xposition_noise", 0.0, 1.0,
                              &pargs->xposition_noise);
    gwy_synth_load_arg_double(settings, key, "yperiod", 1.0, 2000.0,
                              &pargs->yperiod);
    gwy_synth_load_arg_double(settings, key, "yposition_noise", 0.0, 1.0,
                              &pargs->yposition_noise);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_dblstair(gpointer p,
                   GwyContainer *settings)
{
    PatSynthArgsDblstair *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/dblstair/");
    gwy_synth_save_arg_double(settings, key, "xperiod", pargs->xperiod);
    gwy_synth_save_arg_double(settings, key, "xposition_noise",
                              pargs->xposition_noise);
    gwy_synth_save_arg_double(settings, key, "yperiod", pargs->yperiod);
    gwy_synth_save_arg_double(settings, key, "yposition_noise",
                              pargs->yposition_noise);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    g_string_free(key, TRUE);
}

static void
reset_dblstair(gpointer p)
{
    PatSynthControlsDblstair *pcontrols = p;
    PatSynthArgsDblstair *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_dblstair;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xperiod),
                             pargs->xperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xposition_noise),
                             pargs->xposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yperiod),
                             pargs->yperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yposition_noise),
                             pargs->yposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
}

/***************************************************************************
 *
 * Grating
 *
 ***************************************************************************/

typedef struct {
    gdouble period;
    gdouble position_noise;
    gdouble top_frac;
    gdouble top_frac_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble asymm;
    gdouble height;
    gdouble height_noise;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
    gboolean scale_with_width;
} PatSynthArgsGrating;

typedef struct {
    PatSynthArgsGrating *args;
    GtkObject *period;
    GtkWidget *period_units;
    GtkObject *position_noise;
    GtkWidget *scale_with_width;
    GtkObject *top_frac;
    GtkWidget *top_frac_units;
    GtkObject *top_frac_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *asymm;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsGrating;

static const PatSynthArgsGrating pat_synth_defaults_grating = {
    50.0, 0.0,
    0.45, 0.0,
    0.05, 0.0,
    0.0,
    1.0, 0.0,
    0.0,
    0.0, 10.0,
    FALSE,
};

static gpointer
create_gui_grating(PatSynthControls *controls)
{
    PatSynthControlsGrating *pcontrols;
    PatSynthArgsGrating *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 5+4+5+3);
    pcontrols = g_new0(PatSynthControlsGrating, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    row = gwy_synth_attach_heading(controls, row, _("Period"), NULL);
    pcontrols->period = gtk_adjustment_new(pargs->period,
                                           1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->period, &pargs->period,
                                   _("_Period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->period_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->position_noise,
                                             &pargs->position_noise);
    row = gwy_synth_attach_checkbox(controls, row,
                                    &pcontrols->scale_with_width,
                                    &pargs->scale_with_width,
                                    _("Scale features with _width"));

    row = gwy_synth_attach_heading(controls, row, _("Duty Cycle"), NULL);
    pcontrols->top_frac = gtk_adjustment_new(pargs->top_frac,
                                             0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->top_frac, &pargs->top_frac,
                              _("_Top fraction:"), GWY_HSCALE_LINEAR,
                              pcontrols->period, NULL,
                              &pcontrols->top_frac_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->top_frac_noise,
                                    &pargs->top_frac_noise);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->slope, &pargs->slope,
                              _("_Slope width:"), GWY_HSCALE_SQRT,
                              pcontrols->period, NULL,
                              &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);
    row = attach_asymmetry(controls, row, &pcontrols->asymm, &pargs->asymm);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_grating(PatSynthControls *controls)
{
    PatSynthControlsGrating *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->period_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->top_frac_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->period));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->top_frac));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_grating(const PatSynthArgs *args,
                     const GwyDimensionArgs *dimsargs,
                     GwyRandGenSet *rngset,
                     GwyDataField *dfield)
{
    const PatSynthArgsGrating *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    gdouble *abscissa, *height, *width, *slopeleft, *sloperight;
    GwyDataField *displx, *disply, *tmap;
    guint n;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    tmap = displx;  /* Alias for clarity. */
    displacement_to_t_linear(tmap, displx, disply, pargs->angle, pargs->period);
    n = find_t_range(tmap, FALSE);
    abscissa = make_positions_1d_linear(n, pargs->position_noise,
                                        rngset, RNG_OFFSET_X);
    width = make_values_1d(n, pargs->top_frac, pargs->top_frac_noise,
                            rngset, RNG_TOP_X);
    height = make_values_1d(n, hmean, pargs->height_noise,
                            rngset, RNG_HEIGHT);
    slopeleft = make_values_1d(n, pargs->slope, pargs->slope_noise,
                               rngset, RNG_SLOPE);
    sloperight = distribute_left_to_left_and_right(slopeleft, n, pargs->asymm);
    if (pargs->scale_with_width) {
        transform_to_scaled_grating(abscissa, width, slopeleft, sloperight, n,
                                    FALSE);
    }
    render_grating(dfield, tmap, abscissa, width, height, slopeleft, sloperight,
                   1, n);

    g_free(slopeleft);
    g_free(sloperight);
    g_free(width);
    g_free(height);
    g_free(abscissa);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_grating(GwyContainer *settings)
{
    PatSynthArgsGrating *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_grating,
                     sizeof(PatSynthArgsGrating));
    key = g_string_new(prefix);
    g_string_append(key, "/grating/");
    gwy_synth_load_arg_double(settings, key, "period", 1.0, 2000.0,
                              &pargs->period);
    gwy_synth_load_arg_double(settings, key, "position_noise", 0.0, 1.0,
                              &pargs->position_noise);
    gwy_synth_load_arg_double(settings, key, "top_frac", 0.0, 1.0,
                              &pargs->top_frac);
    gwy_synth_load_arg_double(settings, key, "top_frac_noise", 0.0, 1.0,
                              &pargs->top_frac_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "asymm", -1.0, 1.0,
                              &pargs->asymm);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    gwy_synth_load_arg_boolean(settings, key, "scale_with_width",
                               &pargs->scale_with_width);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_grating(gpointer p,
                  GwyContainer *settings)
{
    PatSynthArgsGrating *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/grating/");
    gwy_synth_save_arg_double(settings, key, "period", pargs->period);
    gwy_synth_save_arg_double(settings, key, "position_noise",
                              pargs->position_noise);
    gwy_synth_save_arg_double(settings, key, "top_frac", pargs->top_frac);
    gwy_synth_save_arg_double(settings, key, "top_frac_noise",
                              pargs->top_frac_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "asymm", pargs->asymm);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    gwy_synth_save_arg_boolean(settings, key, "scale_with_width",
                               pargs->scale_with_width);
    g_string_free(key, TRUE);
}

static void
reset_grating(gpointer p)
{
    PatSynthControlsGrating *pcontrols = p;
    PatSynthArgsGrating *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_grating;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->period), pargs->period);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->position_noise),
                             pargs->position_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac),
                             pargs->top_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac_noise),
                             pargs->top_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->asymm), pargs->asymm);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pcontrols->scale_with_width),
                                 pargs->scale_with_width);
}

/***************************************************************************
 *
 * Holes (rectangular)
 *
 ***************************************************************************/

typedef struct {
    gdouble xperiod;
    gdouble xposition_noise;
    gdouble yperiod;
    gdouble yposition_noise;
    gdouble xtop_frac;
    gdouble xtop_frac_noise;
    gdouble ytop_frac;
    gdouble ytop_frac_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble roundness;
    gdouble roundness_noise;
    gdouble height;
    gdouble height_noise;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
} PatSynthArgsRHoles;

typedef struct {
    PatSynthArgsRHoles *args;
    GtkObject *xperiod;
    GtkWidget *xperiod_units;
    GtkObject *xposition_noise;
    GtkObject *yperiod;
    GtkWidget *yperiod_units;
    GtkObject *yposition_noise;
    GtkObject *xtop_frac;
    GtkWidget *xtop_frac_units;
    GtkObject *xtop_frac_noise;
    GtkObject *ytop_frac;
    GtkWidget *ytop_frac_units;
    GtkObject *ytop_frac_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *roundness;
    GtkObject *roundness_noise;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsRHoles;

static const PatSynthArgsRHoles pat_synth_defaults_rholes = {
    50.0, 0.0,
    50.0, 0.0,
    0.3, 0.0,
    0.3, 0.0,
    0.05, 0.0,
    0.1, 0.0,
    1.0, 0.0,
    0.0,
    0.0, 10.0,
};

static gpointer
create_gui_rholes(PatSynthControls *controls)
{
    PatSynthControlsRHoles *pcontrols;
    PatSynthArgsRHoles *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 7+7+3+3+3);
    pcontrols = g_new0(PatSynthControlsRHoles, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    row = gwy_synth_attach_heading(controls, row, _("Period"), NULL);
    pcontrols->xperiod = gtk_adjustment_new(pargs->xperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->xperiod, &pargs->xperiod,
                                   _("_X period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->xperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->xposition_noise,
                                             &pargs->xposition_noise);
    pcontrols->yperiod = gtk_adjustment_new(pargs->yperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->yperiod, &pargs->yperiod,
                                   _("_Y period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->yperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->yposition_noise,
                                             &pargs->yposition_noise);

    row = gwy_synth_attach_heading(controls, row, _("Duty Cycle"), NULL);
    pcontrols->xtop_frac = gtk_adjustment_new(pargs->xtop_frac,
                                             0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->xtop_frac, &pargs->xtop_frac,
                              _("X top fraction:"), GWY_HSCALE_LINEAR,
                              pcontrols->xperiod, NULL,
                              &pcontrols->xtop_frac_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->xtop_frac_noise,
                                    &pargs->xtop_frac_noise);
    pcontrols->ytop_frac = gtk_adjustment_new(pargs->ytop_frac,
                                             0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->ytop_frac, &pargs->ytop_frac,
                              _("Y top fraction:"), GWY_HSCALE_LINEAR,
                              pcontrols->yperiod, NULL,
                              &pcontrols->ytop_frac_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->ytop_frac_noise,
                                    &pargs->ytop_frac_noise);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac_of_min(controls, row,
                                     pcontrols->slope, &pargs->slope,
                                     _("_Slope width:"), GWY_HSCALE_SQRT,
                                     pcontrols->xperiod, pcontrols->yperiod,
                                     NULL, &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);

    row = gwy_synth_attach_roundness(controls, row,
                                     &pcontrols->roundness, &pargs->roundness);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->roundness_noise,
                                    &pargs->roundness_noise);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_rholes(PatSynthControls *controls)
{
    PatSynthControlsRHoles *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->xperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->yperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->xtop_frac_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->ytop_frac_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->xperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->yperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->xtop_frac));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->ytop_frac));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_rholes(const PatSynthArgs *args,
                    const GwyDimensionArgs *dimsargs,
                    GwyRandGenSet *rngset,
                    GwyDataField *dfield)
{
    const PatSynthArgsRHoles *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    gdouble *abscissau, *abscissav, *height, *xsize, *ysize, *slope, *roundness;
    GwyDataField *displx, *disply, *umap, *vmap;
    guint n, nu, nv;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    umap = displx;  /* Alias for clarity. */
    vmap = disply;  /* Alias for clarity. */
    displacement_to_uv_linear(umap, vmap, displx, disply, pargs->angle,
                              pargs->xperiod, pargs->yperiod);
    nu = find_t_range(umap, FALSE);
    nv = find_t_range(vmap, FALSE);
    n = MAX(nu, nv);
    abscissau = make_positions_2d_linear(n, pargs->xposition_noise, FALSE,
                                         rngset, RNG_OFFSET_X);
    abscissav = make_positions_2d_linear(n, pargs->yposition_noise, TRUE,
                                         rngset, RNG_OFFSET_Y);
    xsize = make_values_2d(n, 1.0 - pargs->xtop_frac, pargs->xtop_frac_noise,
                           rngset, RNG_TOP_X);
    ysize = make_values_2d(n, 1.0 - pargs->ytop_frac, pargs->ytop_frac_noise,
                           rngset, RNG_TOP_Y);
    slope = make_values_2d(n, pargs->slope, pargs->slope_noise,
                           rngset, RNG_SLOPE);
    roundness = make_values_2d(n, pargs->roundness, pargs->roundness_noise,
                               rngset, RNG_ROUNDNESS);
    height = make_values_2d(n, hmean, pargs->height_noise,
                            rngset, RNG_HEIGHT);
    render_holes(dfield, umap, vmap, abscissau, abscissav,
                 xsize, ysize, slope, roundness,
                 height, n, pargs->yperiod/pargs->xperiod);

    g_free(height);
    g_free(roundness);
    g_free(slope);
    g_free(ysize);
    g_free(xsize);
    g_free(abscissau);
    g_free(abscissav);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_rholes(GwyContainer *settings)
{
    PatSynthArgsRHoles *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_rholes, sizeof(PatSynthArgsRHoles));
    key = g_string_new(prefix);
    g_string_append(key, "/rholes/");
    gwy_synth_load_arg_double(settings, key, "xperiod", 1.0, 2000.0,
                              &pargs->xperiod);
    gwy_synth_load_arg_double(settings, key, "xposition_noise", 0.0, 1.0,
                              &pargs->xposition_noise);
    gwy_synth_load_arg_double(settings, key, "yperiod", 1.0, 2000.0,
                              &pargs->yperiod);
    gwy_synth_load_arg_double(settings, key, "yposition_noise", 0.0, 1.0,
                              &pargs->yposition_noise);
    gwy_synth_load_arg_double(settings, key, "xtop_frac", 0.0, 1.0,
                              &pargs->xtop_frac);
    gwy_synth_load_arg_double(settings, key, "xtop_frac_noise", 0.0, 1.0,
                              &pargs->xtop_frac_noise);
    gwy_synth_load_arg_double(settings, key, "ytop_frac", 0.0, 1.0,
                              &pargs->ytop_frac);
    gwy_synth_load_arg_double(settings, key, "ytop_frac_noise", 0.0, 1.0,
                              &pargs->ytop_frac_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "roundness", 0.0, 1.0,
                              &pargs->roundness);
    gwy_synth_load_arg_double(settings, key, "roundness_noise", 0.0, 1.0,
                              &pargs->roundness_noise);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_rholes(gpointer p,
                 GwyContainer *settings)
{
    PatSynthArgsRHoles *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/rholes/");
    gwy_synth_save_arg_double(settings, key, "xperiod", pargs->xperiod);
    gwy_synth_save_arg_double(settings, key, "xposition_noise",
                              pargs->xposition_noise);
    gwy_synth_save_arg_double(settings, key, "yperiod", pargs->yperiod);
    gwy_synth_save_arg_double(settings, key, "yposition_noise",
                              pargs->yposition_noise);
    gwy_synth_save_arg_double(settings, key, "xtop_frac", pargs->xtop_frac);
    gwy_synth_save_arg_double(settings, key, "xtop_frac_noise",
                              pargs->xtop_frac_noise);
    gwy_synth_save_arg_double(settings, key, "ytop_frac", pargs->ytop_frac);
    gwy_synth_save_arg_double(settings, key, "ytop_frac_noise",
                              pargs->ytop_frac_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "roundness", pargs->roundness);
    gwy_synth_save_arg_double(settings, key, "roundness_noise",
                              pargs->roundness_noise);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    g_string_free(key, TRUE);
}

static void
reset_rholes(gpointer p)
{
    PatSynthControlsRHoles *pcontrols = p;
    PatSynthArgsRHoles *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_rholes;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xperiod),
                             pargs->xperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xposition_noise),
                             pargs->xposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yperiod),
                             pargs->yperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yposition_noise),
                             pargs->yposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xtop_frac),
                             pargs->xtop_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xtop_frac_noise),
                             pargs->xtop_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->ytop_frac),
                             pargs->ytop_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->ytop_frac_noise),
                             pargs->ytop_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->roundness),
                             pargs->roundness);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->roundness_noise),
                             pargs->roundness_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
}

/***************************************************************************
 *
 * Pillars
 *
 ***************************************************************************/

typedef struct {
    gdouble xperiod;
    gdouble xposition_noise;
    gdouble yperiod;
    gdouble yposition_noise;
    gdouble size_frac;
    gdouble size_frac_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble orientation;
    gdouble orientation_noise;
    gdouble height;
    gdouble height_noise;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
    PillarShapeType shape;
} PatSynthArgsPillars;

typedef struct {
    PatSynthArgsPillars *args;
    GtkWidget *shape;
    GtkObject *xperiod;
    GtkWidget *xperiod_units;
    GtkObject *xposition_noise;
    GtkObject *yperiod;
    GtkWidget *yperiod_units;
    GtkObject *yposition_noise;
    GtkObject *size_frac;
    GtkWidget *size_frac_units;
    GtkObject *size_frac_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *orientation;
    GtkObject *orientation_noise;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsPillars;

static const PatSynthArgsPillars pat_synth_defaults_pillars = {
    50.0, 0.0,
    50.0, 0.0,
    0.5, 0.0,
    0.05, 0.0,
    0.0, 0.0,
    1.0, 0.0,
    0.0,
    0.0, 10.0,
    PILLAR_SHAPE_CIRCLE,
};

static void
pillar_shape_selected(GtkComboBox *combo, PatSynthControls *controls)
{
    PatSynthControlsPillars *pcontrols = controls->pattern_controls;
    PatSynthArgsPillars *pargs = pcontrols->args;

    pargs->shape = gwy_enum_combo_box_get_active(combo);
    pat_synth_invalidate(controls);
}

static gpointer
create_gui_pillars(PatSynthControls *controls)
{
    static const GwyEnum shapes[] = {
        { N_("Circle"),  PILLAR_SHAPE_CIRCLE,  },
        { N_("Square"),  PILLAR_SHAPE_SQUARE,  },
        { N_("Hexagon"), PILLAR_SHAPE_HEXAGON, },
    };

    PatSynthControlsPillars *pcontrols;
    PatSynthArgsPillars *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 1+7+4+4+4+2);
    pcontrols = g_new0(PatSynthControlsPillars, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    gtk_table_set_row_spacing(controls->table, row-1, 8);
    pcontrols->shape
        = gwy_enum_combo_box_new(shapes, PILLAR_SHAPE_NSHAPES,
                                 G_CALLBACK(pillar_shape_selected), controls,
                                 pargs->shape, TRUE);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("S_hape:"), NULL,
                            GTK_OBJECT(pcontrols->shape),
                            GWY_HSCALE_WIDGET_NO_EXPAND);

    row = gwy_synth_attach_heading(controls, row, _("Period"), NULL);
    pcontrols->xperiod = gtk_adjustment_new(pargs->xperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->xperiod, &pargs->xperiod,
                                   _("_X period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->xperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->xposition_noise,
                                             &pargs->xposition_noise);
    pcontrols->yperiod = gtk_adjustment_new(pargs->yperiod,
                                            1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->yperiod, &pargs->yperiod,
                                   _("_Y period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->yperiod_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->yposition_noise,
                                             &pargs->yposition_noise);

    row = gwy_synth_attach_heading(controls, row, _("Size"), NULL);
    pcontrols->size_frac = gtk_adjustment_new(pargs->size_frac,
                                             0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac_of_min(controls, row,
                                     pcontrols->size_frac, &pargs->size_frac,
                                     _("Si_ze:"), GWY_HSCALE_LINEAR,
                                     pcontrols->xperiod, pcontrols->yperiod,
                                     NULL, &pcontrols->size_frac_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->size_frac_noise,
                                    &pargs->size_frac_noise);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac_of_min(controls, row,
                                     pcontrols->slope, &pargs->slope,
                                     _("_Slope width:"), GWY_HSCALE_SQRT,
                                     pcontrols->xperiod, pcontrols->yperiod,
                                     NULL, &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->orientation,
                                       &pargs->orientation);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->orientation_noise,
                                    &pargs->orientation_noise);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_pillars(PatSynthControls *controls)
{
    PatSynthControlsPillars *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->xperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->yperiod_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->size_frac_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->xperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->yperiod));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->size_frac));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_pillars(const PatSynthArgs *args,
                     const GwyDimensionArgs *dimsargs,
                     GwyRandGenSet *rngset,
                     GwyDataField *dfield)
{
    const PatSynthArgsPillars *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    gdouble *abscissau, *abscissav, *height, *size, *slope, *sine, *cosine;
    GwyDataField *displx, *disply, *umap, *vmap;
    gdouble orinoise = 0.0;
    guint n, nu, nv;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    umap = displx;  /* Alias for clarity. */
    vmap = disply;  /* Alias for clarity. */
    displacement_to_uv_linear(umap, vmap, displx, disply, pargs->angle,
                              pargs->xperiod, pargs->yperiod);
    nu = find_t_range(umap, FALSE);
    nv = find_t_range(vmap, FALSE);
    n = MAX(nu, nv);
    abscissau = make_positions_2d_linear(n, pargs->xposition_noise, FALSE,
                                         rngset, RNG_OFFSET_X);
    abscissav = make_positions_2d_linear(n, pargs->yposition_noise, TRUE,
                                         rngset, RNG_OFFSET_Y);
    size = make_values_2d(n, pargs->size_frac, pargs->size_frac_noise,
                          rngset, RNG_SIZE_X);
    slope = make_values_2d(n, pargs->slope, pargs->slope_noise,
                           rngset, RNG_SLOPE);
    if (pargs->shape == PILLAR_SHAPE_SQUARE)
        orinoise = G_PI/4.0 * pargs->orientation_noise;
    else if (pargs->shape == PILLAR_SHAPE_HEXAGON)
        orinoise = G_PI/6.0 * pargs->orientation_noise;
    sine = make_values_2d_gaussian(n, pargs->orientation, orinoise,
                                   rngset, RNG_ORIENTATION);
    cosine = g_new(gdouble, n*n);
    transform_to_sine_cosine(sine, cosine, n*n);
    height = make_values_2d(n, hmean, pargs->height_noise,
                            rngset, RNG_HEIGHT);
    render_pillars(dfield, umap, vmap, pargs->shape, abscissau, abscissav,
                   size, slope, sine, cosine,
                   height, n, pargs->yperiod/pargs->xperiod);

    g_free(height);
    g_free(sine);
    g_free(cosine);
    g_free(slope);
    g_free(size);
    g_free(abscissau);
    g_free(abscissav);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_pillars(GwyContainer *settings)
{
    PatSynthArgsPillars *pargs;
    GString *key;
    gint32 i;

    pargs = g_memdup(&pat_synth_defaults_pillars, sizeof(PatSynthArgsPillars));
    key = g_string_new(prefix);
    g_string_append(key, "/pillars/");
    i = pargs->shape;
    gwy_synth_load_arg_int32(settings, key, "shape", 0, PILLAR_SHAPE_NSHAPES-1,
                             &i);
    pargs->shape = i;
    gwy_synth_load_arg_double(settings, key, "xperiod", 1.0, 2000.0,
                              &pargs->xperiod);
    gwy_synth_load_arg_double(settings, key, "xposition_noise", 0.0, 1.0,
                              &pargs->xposition_noise);
    gwy_synth_load_arg_double(settings, key, "yperiod", 1.0, 2000.0,
                              &pargs->yperiod);
    gwy_synth_load_arg_double(settings, key, "yposition_noise", 0.0, 1.0,
                              &pargs->yposition_noise);
    gwy_synth_load_arg_double(settings, key, "size_frac", 0.0, 1.0,
                              &pargs->size_frac);
    gwy_synth_load_arg_double(settings, key, "size_frac_noise", 0.0, 1.0,
                              &pargs->size_frac_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "orientation", -G_PI, G_PI,
                              &pargs->orientation);
    gwy_synth_load_arg_double(settings, key, "orientation_noise", 0.0, 1.0,
                              &pargs->orientation_noise);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_pillars(gpointer p,
                  GwyContainer *settings)
{
    PatSynthArgsPillars *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/pillars/");
    gwy_synth_save_arg_int32(settings, key, "shape", pargs->shape);
    gwy_synth_save_arg_double(settings, key, "xperiod", pargs->xperiod);
    gwy_synth_save_arg_double(settings, key, "xposition_noise",
                              pargs->xposition_noise);
    gwy_synth_save_arg_double(settings, key, "yperiod", pargs->yperiod);
    gwy_synth_save_arg_double(settings, key, "yposition_noise",
                              pargs->yposition_noise);
    gwy_synth_save_arg_double(settings, key, "size_frac", pargs->size_frac);
    gwy_synth_save_arg_double(settings, key, "size_frac_noise",
                              pargs->size_frac_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "orientation", pargs->orientation);
    gwy_synth_save_arg_double(settings, key, "orientation_noise",
                              pargs->orientation_noise);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    g_string_free(key, TRUE);
}

static void
reset_pillars(gpointer p)
{
    PatSynthControlsPillars *pcontrols = p;
    PatSynthArgsPillars *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_pillars;

    gwy_enum_combo_box_set_active(GTK_COMBO_BOX(pcontrols->shape),
                                  pargs->shape);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xperiod),
                             pargs->xperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xposition_noise),
                             pargs->xposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yperiod),
                             pargs->yperiod);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->yposition_noise),
                             pargs->yposition_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->size_frac),
                             pargs->size_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->size_frac_noise),
                             pargs->size_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->orientation),
                             pargs->orientation);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->orientation_noise),
                             pargs->orientation_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
}

/***************************************************************************
 *
 * Amphitheatre
 *
 ***************************************************************************/

typedef struct {
    gdouble invpower;
    gdouble flat;
    gdouble position_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble height;
    gdouble height_noise;
    gdouble parabolicity;
    gdouble xcenter;
    gdouble ycenter;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
} PatSynthArgsAmphith;

typedef struct {
    PatSynthArgsAmphith *args;
    GtkObject *invpower;
    GtkObject *flat;
    GtkWidget *flat_units;
    GtkObject *position_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *parabolicity;
    GtkObject *xcenter;
    GtkObject *ycenter;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_value;
    GtkWidget *tau_units;
} PatSynthControlsAmphith;

static const PatSynthArgsAmphith pat_synth_defaults_amphith = {
    1.0,
    50.0, 0.0,
    0.05, 0.0,
    1.0, 0.0,
    0.0,
    0.0, 0.0,
    0.0,
    0.0, 10.0,
};

static gpointer
create_gui_amphith(PatSynthControls *controls)
{
    PatSynthControlsAmphith *pcontrols;
    PatSynthArgsAmphith *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 1+5+4+3);
    pcontrols = g_new0(PatSynthControlsAmphith, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    gtk_table_set_row_spacing(controls->table, row-1, 8);

    row = attach_superellipse_invpower(controls, row,
                                       &pcontrols->invpower, &pargs->invpower);

    row = gwy_synth_attach_heading(controls, row, _("Terrace"), NULL);
    pcontrols->flat = gtk_adjustment_new(pargs->flat,
                                         1.0, 1000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->flat, &pargs->flat,
                                   _("_Terrace width:"), GWY_HSCALE_LOG,
                                   NULL,
                                   NULL, &pcontrols->flat_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->position_noise,
                                             &pargs->position_noise);

    row = attach_parabolicity(controls, row,
                              &pcontrols->parabolicity, &pargs->parabolicity);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->slope, &pargs->slope,
                              _("_Slope width:"), GWY_HSCALE_SQRT,
                              pcontrols->flat, NULL,
                              &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    row = gwy_synth_attach_heading(controls, row, _("Position"), NULL);
    row = attach_centre_offset(controls, row,
                               &pcontrols->xcenter, &pargs->xcenter, FALSE);
    row = attach_centre_offset(controls, row,
                               &pcontrols->ycenter, &pargs->ycenter, TRUE);

    return pcontrols;
}

static void
dimensions_changed_amphith(PatSynthControls *controls)
{
    PatSynthControlsAmphith *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->flat_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->flat));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_amphith(const PatSynthArgs *args,
                     const GwyDimensionArgs *dimsargs,
                     GwyRandGenSet *rngset,
                     GwyDataField *dfield)
{
    const PatSynthArgsAmphith *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    GwyDataField *displx, *disply, *tmap;
    gdouble *radius, *height, *slope;
    gdouble scale;
    guint n;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    tmap = displx;  /* Alias for clarity. */
    displacement_to_t_superellipse(tmap, displx, disply,
                                   pargs->angle, pargs->xcenter, pargs->ycenter,
                                   pargs->invpower, pargs->flat);
    n = find_t_range(tmap, TRUE);
    scale = 0.5*hypot(xres, yres)/pargs->flat;
    radius = make_positions_1d_radial(n, pargs->position_noise,
                                      scale, pargs->parabolicity,
                                      rngset, RNG_OFFSET_X);
    height = make_heights_staircase(radius, n,
                                    hmean, pargs->height_noise,
                                    FALSE, TRUE,
                                    rngset, RNG_HEIGHT);
    slope = make_values_1d(n, pargs->slope, pargs->slope_noise,
                           rngset, RNG_SLOPE);
    render_staircase(dfield, tmap, radius, height, slope, n);

    g_free(slope);
    g_free(height);
    g_free(radius);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_amphith(GwyContainer *settings)
{
    PatSynthArgsAmphith *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_amphith, sizeof(PatSynthArgsAmphith));
    key = g_string_new(prefix);
    g_string_append(key, "/amphith/");
    gwy_synth_load_arg_double(settings, key, "flat", 1.0, 1000.0,
                              &pargs->flat);
    gwy_synth_load_arg_double(settings, key, "position_noise", 0.0, 1.0,
                              &pargs->position_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "invpower", 0.0, 2.0,
                              &pargs->invpower);
    gwy_synth_load_arg_double(settings, key, "parabolicity", -1.0, 1.0,
                              &pargs->parabolicity);
    gwy_synth_load_arg_double(settings, key, "xcenter", -2.0, 2.0,
                              &pargs->xcenter);
    gwy_synth_load_arg_double(settings, key, "ycenter", -2.0, 2.0,
                              &pargs->ycenter);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_amphith(gpointer p,
                GwyContainer *settings)
{
    PatSynthArgsAmphith *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/amphith/");
    gwy_synth_save_arg_double(settings, key, "flat", pargs->flat);
    gwy_synth_save_arg_double(settings, key, "position_noise",
                              pargs->position_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "invpower", pargs->invpower);
    gwy_synth_save_arg_double(settings, key, "parabolicity",
                              pargs->parabolicity);
    gwy_synth_save_arg_double(settings, key, "xcenter", pargs->xcenter);
    gwy_synth_save_arg_double(settings, key, "ycenter", pargs->ycenter);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    g_string_free(key, TRUE);
}

static void
reset_amphith(gpointer p)
{
    PatSynthControlsAmphith *pcontrols = p;
    PatSynthArgsAmphith *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_amphith;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->flat), pargs->flat);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->position_noise),
                             pargs->position_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xcenter),
                             pargs->xcenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->ycenter),
                             pargs->ycenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->parabolicity),
                             pargs->parabolicity);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->invpower),
                             pargs->invpower);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
}

/***************************************************************************
 *
 * Concentric rings
 *
 ***************************************************************************/

typedef struct {
    gdouble invpower;
    gdouble period;
    gdouble position_noise;
    gdouble top_frac;
    gdouble top_frac_noise;
    gdouble slope;
    gdouble slope_noise;
    gdouble asymm;
    gdouble height;
    gdouble height_noise;
    gdouble parabolicity;
    gdouble xcenter;
    gdouble ycenter;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
    gboolean scale_with_width;
} PatSynthArgsRings;

typedef struct {
    PatSynthArgsRings *args;
    GtkObject *invpower;
    GtkObject *period;
    GtkWidget *period_units;
    GtkObject *position_noise;
    GtkWidget *scale_with_width;
    GtkObject *top_frac;
    GtkWidget *top_frac_units;
    GtkObject *top_frac_noise;
    GtkObject *slope;
    GtkWidget *slope_units;
    GtkObject *slope_noise;
    GtkObject *asymm;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *height_noise;
    GtkObject *parabolicity;
    GtkObject *xcenter;
    GtkObject *ycenter;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_value;
    GtkWidget *tau_units;
} PatSynthControlsRings;

static const PatSynthArgsRings pat_synth_defaults_rings = {
    1.0,
    50.0, 0.0,
    0.5, 0.0,
    0.05, 0.0,
    0.0,
    1.0, 0.0,
    0.0,
    0.0, 0.0,
    0.0,
    0.0, 10.0,
    FALSE,
};

static gpointer
create_gui_rings(PatSynthControls *controls)
{
    PatSynthControlsRings *pcontrols;
    PatSynthArgsRings *pargs;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 1+6+4+5+3);
    pcontrols = g_new0(PatSynthControlsRings, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    gtk_table_set_row_spacing(controls->table, row-1, 8);

    row = attach_superellipse_invpower(controls, row,
                                       &pcontrols->invpower, &pargs->invpower);

    row = gwy_synth_attach_heading(controls, row, _("Period"), NULL);
    pcontrols->period = gtk_adjustment_new(pargs->period,
                                           1.0, 2000.0, 0.01, 10.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->period, &pargs->period,
                                   _("_Period:"), GWY_HSCALE_LOG,
                                   NULL, NULL, &pcontrols->period_units);
    row = gwy_synth_attach_position_variance(controls, row,
                                             &pcontrols->position_noise,
                                             &pargs->position_noise);
    row = attach_parabolicity(controls, row,
                              &pcontrols->parabolicity, &pargs->parabolicity);
    row = gwy_synth_attach_checkbox(controls, row,
                                    &pcontrols->scale_with_width,
                                    &pargs->scale_with_width,
                                    _("Scale features with _width"));

    row = gwy_synth_attach_heading(controls, row, _("Duty Cycle"), NULL);
    pcontrols->top_frac = gtk_adjustment_new(pargs->top_frac,
                                             0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->top_frac, &pargs->top_frac,
                              _("_Top fraction:"), GWY_HSCALE_LINEAR,
                              pcontrols->period, NULL,
                              &pcontrols->top_frac_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->top_frac_noise,
                                    &pargs->top_frac_noise);

    row = gwy_synth_attach_heading(controls, row, _("Slope"), NULL);
    pcontrols->slope = gtk_adjustment_new(pargs->slope,
                                          0.0, 1.0, 0.001, 0.1, 0);
    row = attach_lateral_frac(controls, row,
                              pcontrols->slope, &pargs->slope,
                              _("_Slope width:"), GWY_HSCALE_SQRT,
                              pcontrols->period, NULL,
                              &pcontrols->slope_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->slope_noise,
                                    &pargs->slope_noise);
    row = attach_asymmetry(controls, row, &pcontrols->asymm, &pargs->asymm);

    row = gwy_synth_attach_heading(controls, row, _("Height"), NULL);
    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);
    row = gwy_synth_attach_variance(controls, row,
                                    &pcontrols->height_noise,
                                    &pargs->height_noise);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 2+3+3);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    row = gwy_synth_attach_heading(controls, row, _("Position"), NULL);
    row = attach_centre_offset(controls, row,
                               &pcontrols->xcenter, &pargs->xcenter, FALSE);
    row = attach_centre_offset(controls, row,
                               &pcontrols->ycenter, &pargs->ycenter, TRUE);

    return pcontrols;
}

static void
dimensions_changed_rings(PatSynthControls *controls)
{
    PatSynthControlsRings *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->period_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->top_frac_units),
                         dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->slope_units), dims->sxyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->period));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->top_frac));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->slope));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static void
make_pattern_rings(const PatSynthArgs *args,
                     const GwyDimensionArgs *dimsargs,
                     GwyRandGenSet *rngset,
                     GwyDataField *dfield)
{
    const PatSynthArgsRings *pargs = args->pattern_args;
    gdouble hmean = pargs->height * pow10(dimsargs->zpow10);
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    GwyDataField *displx, *disply, *tmap;
    gdouble *radius, *width, *height, *slopeleft, *sloperight;
    gdouble scale;
    guint n;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);
    tmap = displx;  /* Alias for clarity. */
    displacement_to_t_superellipse(tmap, displx, disply,
                                   pargs->angle, pargs->xcenter, pargs->ycenter,
                                   pargs->invpower, pargs->period);
    n = find_t_range(tmap, TRUE);
    scale = 0.5*hypot(xres, yres)/pargs->period;
    radius = make_positions_1d_radial(n, pargs->position_noise,
                                      scale, pargs->parabolicity,
                                      rngset, RNG_OFFSET_X);
    width = make_values_1d(n, pargs->top_frac, pargs->top_frac_noise,
                            rngset, RNG_TOP_X);
    height = make_values_1d(n, hmean, pargs->height_noise,
                            rngset, RNG_HEIGHT);
    slopeleft = make_values_1d(n, pargs->slope, pargs->slope_noise,
                               rngset, RNG_SLOPE);
    sloperight = distribute_left_to_left_and_right(slopeleft, n, pargs->asymm);
    if (pargs->scale_with_width) {
        transform_to_scaled_grating(radius, width, slopeleft, sloperight, n,
                                    TRUE);
    }
    render_grating(dfield, tmap, radius, width, height, slopeleft, sloperight,
                   1, n);

    g_free(sloperight);
    g_free(slopeleft);
    g_free(height);
    g_free(width);
    g_free(radius);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_rings(GwyContainer *settings)
{
    PatSynthArgsRings *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_rings, sizeof(PatSynthArgsRings));
    key = g_string_new(prefix);
    g_string_append(key, "/rings/");
    gwy_synth_load_arg_double(settings, key, "period", 1.0, 1000.0,
                              &pargs->period);
    gwy_synth_load_arg_double(settings, key, "position_noise", 0.0, 1.0,
                              &pargs->position_noise);
    gwy_synth_load_arg_double(settings, key, "top_frac", 0.0, 1.0,
                              &pargs->top_frac);
    gwy_synth_load_arg_double(settings, key, "top_frac_noise", 0.0, 1.0,
                              &pargs->top_frac_noise);
    gwy_synth_load_arg_double(settings, key, "slope", 0.0, 1.0,
                              &pargs->slope);
    gwy_synth_load_arg_double(settings, key, "slope_noise", 0.0, 1.0,
                              &pargs->slope_noise);
    gwy_synth_load_arg_double(settings, key, "asymm", -1.0, 1.0,
                              &pargs->asymm);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "height_noise", 0.0, 1.0,
                              &pargs->height_noise);
    gwy_synth_load_arg_double(settings, key, "invpower", 0.0, 2.0,
                              &pargs->invpower);
    gwy_synth_load_arg_double(settings, key, "parabolicity", -1.0, 1.0,
                              &pargs->parabolicity);
    gwy_synth_load_arg_double(settings, key, "xcenter", -2.0, 2.0,
                              &pargs->xcenter);
    gwy_synth_load_arg_double(settings, key, "ycenter", -2.0, 2.0,
                              &pargs->ycenter);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    gwy_synth_load_arg_boolean(settings, key, "scale_with_width",
                               &pargs->scale_with_width);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_rings(gpointer p,
                GwyContainer *settings)
{
    PatSynthArgsRings *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/rings/");
    gwy_synth_save_arg_double(settings, key, "period", pargs->period);
    gwy_synth_save_arg_double(settings, key, "position_noise",
                              pargs->position_noise);
    gwy_synth_save_arg_double(settings, key, "top_frac", pargs->top_frac);
    gwy_synth_save_arg_double(settings, key, "top_frac_noise",
                              pargs->top_frac_noise);
    gwy_synth_save_arg_double(settings, key, "slope", pargs->slope);
    gwy_synth_save_arg_double(settings, key, "slope_noise", pargs->slope_noise);
    gwy_synth_save_arg_double(settings, key, "asymm", pargs->asymm);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "height_noise",
                              pargs->height_noise);
    gwy_synth_save_arg_double(settings, key, "invpower", pargs->invpower);
    gwy_synth_save_arg_double(settings, key, "parabolicity",
                              pargs->parabolicity);
    gwy_synth_save_arg_double(settings, key, "xcenter", pargs->xcenter);
    gwy_synth_save_arg_double(settings, key, "ycenter", pargs->ycenter);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    gwy_synth_save_arg_boolean(settings, key, "scale_with_width",
                               pargs->scale_with_width);
    g_string_free(key, TRUE);
}

static void
reset_rings(gpointer p)
{
    PatSynthControlsRings *pcontrols = p;
    PatSynthArgsRings *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_rings;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->period), pargs->period);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->position_noise),
                             pargs->position_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac),
                             pargs->top_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac_noise),
                             pargs->top_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope), pargs->slope);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->slope_noise),
                             pargs->slope_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->asymm), pargs->asymm);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height_noise),
                             pargs->height_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xcenter),
                             pargs->xcenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->ycenter),
                             pargs->ycenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->parabolicity),
                             pargs->parabolicity);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->invpower),
                             pargs->invpower);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pcontrols->scale_with_width),
                                 pargs->scale_with_width);
}

/***************************************************************************
 *
 * Siemens star
 *
 ***************************************************************************/

typedef struct {
    gint n_rays;
    gdouble top_frac;
    gdouble top_frac_noise;
    gdouble edge_shift;
    gdouble height;
    gdouble xcenter;
    gdouble ycenter;
    gdouble angle;
    gdouble sigma;
    gdouble tau;
} PatSynthArgsStar;

typedef struct {
    PatSynthArgsStar *args;
    GtkObject *n_rays;
    GtkObject *top_frac;
    GtkObject *top_frac_noise;
    GtkObject *edge_shift;
    GtkWidget *edge_shift_units;
    GtkObject *height;
    GtkWidget *height_units;
    GtkObject *xcenter;
    GtkObject *ycenter;
    GtkObject *angle;
    GtkObject *sigma;
    GtkObject *tau;
    GtkWidget *tau_units;
} PatSynthControlsStar;

static const PatSynthArgsStar pat_synth_defaults_star = {
    8,
    0.5, 0.0,
    0.0,
    1.0,
    0.0, 0.0,
    0.0,
    0.0, 10.0,
};

static gpointer
create_gui_star(PatSynthControls *controls)
{
    PatSynthControlsStar *pcontrols;
    PatSynthArgsStar *pargs;
    GtkWidget *spin;
    gint row;

    controls->table = controls->table_parameters;
    row = gwy_synth_extend_table(controls->table, 4+3+3);
    pcontrols = g_new0(PatSynthControlsStar, 1);
    pargs = pcontrols->args = controls->args->pattern_args;

    gtk_table_set_row_spacing(controls->table, row-1, 8);
    pcontrols->n_rays = gtk_adjustment_new(pargs->n_rays, 2, 36, 1, 2, 0);
    g_object_set_data(G_OBJECT(pcontrols->n_rays), "target", &pargs->n_rays);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("_Number of sectors:"), NULL,
                            GTK_OBJECT(pcontrols->n_rays),
                            GWY_HSCALE_LINEAR | GWY_HSCALE_SNAP);
    g_signal_connect_swapped(pcontrols->n_rays, "value-changed",
                             G_CALLBACK(gwy_synth_int_changed), controls);
    row++;

    row = gwy_synth_attach_heading(controls, row, _("Duty Cycle"), NULL);

    pcontrols->top_frac = gtk_adjustment_new(pargs->top_frac,
                                             0.01, 0.99, 0.01, 0.1, 0);
    g_object_set_data(G_OBJECT(pcontrols->top_frac), "target",
                      &pargs->top_frac);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("_Top fraction:"), NULL,
                            GTK_OBJECT(pcontrols->top_frac),
                            GWY_HSCALE_LINEAR);
    g_signal_connect_swapped(pcontrols->top_frac, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);
    row++;

    pcontrols->top_frac_noise = gtk_adjustment_new(pargs->top_frac_noise,
                                                   0, 0.99, 0.01, 0.1, 0);
    g_object_set_data(G_OBJECT(pcontrols->top_frac_noise), "target",
                      &pargs->top_frac_noise);
    gwy_table_attach_adjbar(GTK_WIDGET(controls->table), row++,
                            _("Variance:"), NULL,
                            GTK_OBJECT(pcontrols->top_frac_noise),
                            GWY_HSCALE_LINEAR);
    g_signal_connect_swapped(pcontrols->top_frac_noise, "value-changed",
                             G_CALLBACK(gwy_synth_double_changed), controls);
    row++;

    row = gwy_synth_attach_heading(controls, row, _("Edge Shift"), NULL);

    pcontrols->edge_shift = gtk_adjustment_new(pargs->edge_shift,
                                               -100.0, 100.0, 0.1, 1.0, 0);
    row = gwy_synth_attach_lateral(controls, row,
                                   pcontrols->edge_shift, &pargs->edge_shift,
                                   _("_Edge shift:"), GWY_HSCALE_SQRT,
                                   &spin, NULL, &pcontrols->edge_shift_units);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 2);

    gtk_table_set_row_spacing(controls->table, row-1, 8);
    gtk_table_attach(controls->table, gwy_label_new_header(_("Height")),
                     0, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    row = gwy_synth_attach_height(controls, row,
                                  &pcontrols->height, &pargs->height,
                                  _("_Height:"),
                                  NULL, &pcontrols->height_units);

    controls->table = controls->table_placement;
    row = gwy_synth_extend_table(controls->table, 3+2+2);

    row = gwy_synth_attach_heading(controls, row, _("Position"), NULL);
    row = attach_centre_offset(controls, row,
                               &pcontrols->xcenter, &pargs->xcenter, FALSE);
    row = attach_centre_offset(controls, row,
                               &pcontrols->ycenter, &pargs->ycenter, TRUE);

    row = gwy_synth_attach_orientation(controls, row,
                                       &pcontrols->angle, &pargs->angle);

    row = gwy_synth_attach_deformation(controls, row,
                                       &pcontrols->sigma, &pargs->sigma,
                                       &pcontrols->tau, &pargs->tau,
                                       NULL, &pcontrols->tau_units);

    return pcontrols;
}

static void
dimensions_changed_star(PatSynthControls *controls)
{
    PatSynthControlsStar *pcontrols = controls->pattern_controls;
    GwyDimensions *dims = controls->dims;

    gtk_label_set_markup(GTK_LABEL(pcontrols->edge_shift_units),
                         dims->xyvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->height_units), dims->zvf->units);
    gtk_label_set_markup(GTK_LABEL(pcontrols->tau_units), dims->sxyvf->units);

    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->edge_shift));
    gwy_synth_update_lateral(controls, GTK_ADJUSTMENT(pcontrols->tau));
}

static inline gboolean
is_within_edge_shift(gdouble x, gdouble y, const gdouble *v0, gdouble eshift)
{
    gdouble e2 = eshift*eshift;
    gdouble d2 = x*x + y*y - e2;
    gdouble dotprod;

    if (d2 < 0.0)
        return TRUE;

    dotprod = x*v0[0] + y*v0[1];
    if (dotprod > 0.0 && d2 < dotprod*dotprod)
        return TRUE;

    dotprod = x*v0[2] + y*v0[3];
    if (dotprod > 0.0 && d2 < dotprod*dotprod)
        return TRUE;

    return FALSE;
}

static void
make_pattern_star(const PatSynthArgs *args,
                  const GwyDimensionArgs *dimsargs,
                  GwyRandGenSet *rngset,
                  GwyDataField *dfield)
{
    const PatSynthArgsStar *pargs = args->pattern_args;
    GwyDataField *displx, *disply;
    guint xres = gwy_data_field_get_xres(dfield);
    guint yres = gwy_data_field_get_yres(dfield);
    guint n, nedge, i;
    gdouble *phiedge, *v0edge, *data, *dx_data, *dy_data;
    gdouble h, c, s, xoff, yoff, eshift, tf, tfnoise;

    displx = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_X);
    disply = make_displacement_map(xres, yres, pargs->sigma, pargs->tau,
                                   rngset, RNG_DISPLAC_Y);

    h = pargs->height * pow10(dimsargs->zpow10);
    n = pargs->n_rays;
    tf = pargs->top_frac;
    tfnoise = pargs->top_frac_noise;
    /* Extend the spoke edge list two back and four forward to ensure we can
     * always safely move forward and backward without doing mod nedge. */
    nedge = 2*n+6;
    v0edge = g_new(gdouble, 2*nedge);
    phiedge = g_new(gdouble, nedge);
    for (i = 0; i < n; i++) {
        gdouble phi = 2.0*G_PI*i/n;
        gdouble width = G_PI/n*tf;
        if (tfnoise) {
            gdouble t = gwy_rand_gen_set_double(rngset, RNG_TOP_X);
            if (t > 0.5)
                width += (2.0*t - 1.0)*(1.0 - tf)*G_PI/n*tfnoise;
            else if (t < 0.5)
                width -= (1.0 - 2.0*t)*tf*G_PI/n*tfnoise;
        }
        phiedge[2*i+2] = phi - width;
        phiedge[2*i+3] = phi + width;
    }
    phiedge[0] = phiedge[nedge-6] - 2.0*G_PI;
    phiedge[1] = phiedge[nedge-5] - 2.0*G_PI;
    phiedge[nedge-4] = phiedge[2] + 2.0*G_PI;
    phiedge[nedge-3] = phiedge[3] + 2.0*G_PI;
    phiedge[nedge-2] = phiedge[4] + 2.0*G_PI;
    phiedge[nedge-1] = phiedge[5] + 2.0*G_PI;
    for (i = 0; i < nedge; i++) {
        v0edge[2*i] = cos(phiedge[i]);
        v0edge[2*i+1] = sin(phiedge[i]);
    }

    xoff = (0.5 + pargs->xcenter)*xres;
    yoff = (0.5 + pargs->ycenter)*yres;
    c = cos(pargs->angle);
    s = sin(pargs->angle);
    eshift = pargs->edge_shift;

    data = gwy_data_field_get_data(dfield);
    dx_data = gwy_data_field_get_data(displx);
    dy_data = gwy_data_field_get_data(disply);

#ifdef _OPENMP
#pragma omp parallel for if (gwy_threads_are_enabled()) default(none) \
            private(i) \
            shared(data,dx_data,dy_data,phiedge,v0edge,xoff,yoff,c,s,h,eshift,xres,yres,nedge)
#endif
    for (i = 0; i < yres; i++) {
        guint j, v, k = nedge/2;

        for (j = 0; j < xres; j++) {
            gdouble x, y, xu, yu, phi;

            xu = (j - xoff)*c - (i - yoff)*s;
            yu = (j - xoff)*s + (i - yoff)*c;
            x = xu + dx_data[i*xres + j];
            y = yu + dy_data[i*xres + j];
            phi = atan2(y, x);
            if (phi < 0.0)
                phi += 2.0*G_PI;
            if (phi < phiedge[k] || phi > phiedge[k+1])
                k = bisect_lower(phiedge, nedge, phi);
            v = !(k & 1);
            /* With non-zero edge shift we might need to change upper or lower
             * (but always only one of them). */
            if (!v
                && eshift > 0.0
                && is_within_edge_shift(x, y, v0edge + 2*k, eshift))
                v = 1;
            else if (v
                     && eshift < 0.0
                     && is_within_edge_shift(x, y, v0edge + 2*k, eshift))
                v = 0;

            data[i*xres + j] = h*v;
        }
    }

    g_free(phiedge);
    g_free(v0edge);
    g_object_unref(displx);
    g_object_unref(disply);
}

static gpointer
load_args_star(GwyContainer *settings)
{
    PatSynthArgsStar *pargs;
    GString *key;

    pargs = g_memdup(&pat_synth_defaults_star, sizeof(PatSynthArgsStar));
    key = g_string_new(prefix);
    g_string_append(key, "/star/");
    gwy_synth_load_arg_int32(settings, key, "n_rays", 2, 32, &pargs->n_rays);
    gwy_synth_load_arg_double(settings, key, "top_frac", 0.01, 0.99,
                              &pargs->top_frac);
    gwy_synth_load_arg_double(settings, key, "top_frac_noise", 0.0, 0.99,
                              &pargs->top_frac_noise);
    gwy_synth_load_arg_double(settings, key, "edge_shift", -100, 100.0,
                              &pargs->edge_shift);
    gwy_synth_load_arg_double(settings, key, "height", 0.0001, 10000.0,
                              &pargs->height);
    gwy_synth_load_arg_double(settings, key, "xcenter", -2.0, 2.0,
                              &pargs->xcenter);
    gwy_synth_load_arg_double(settings, key, "ycenter", -2.0, 2.0,
                              &pargs->ycenter);
    gwy_synth_load_arg_double(settings, key, "angle", -G_PI, G_PI,
                              &pargs->angle);
    gwy_synth_load_arg_double(settings, key, "sigma", 0.0, 100.0,
                              &pargs->sigma);
    gwy_synth_load_arg_double(settings, key, "tau", 0.1, 1000.0,
                              &pargs->tau);
    g_string_free(key, TRUE);

    return pargs;
}

static void
save_args_star(gpointer p,
                GwyContainer *settings)
{
    PatSynthArgsStar *pargs = p;
    GString *key;

    key = g_string_new(prefix);
    g_string_append(key, "/star/");
    gwy_synth_save_arg_int32(settings, key, "n_rays", pargs->n_rays);
    gwy_synth_save_arg_double(settings, key, "top_frac", pargs->top_frac);
    gwy_synth_save_arg_double(settings, key, "top_frac_noise",
                              pargs->top_frac_noise);
    gwy_synth_save_arg_double(settings, key, "edge_shift", pargs->edge_shift);
    gwy_synth_save_arg_double(settings, key, "height", pargs->height);
    gwy_synth_save_arg_double(settings, key, "xcenter", pargs->xcenter);
    gwy_synth_save_arg_double(settings, key, "ycenter", pargs->ycenter);
    gwy_synth_save_arg_double(settings, key, "angle", pargs->angle);
    gwy_synth_save_arg_double(settings, key, "sigma", pargs->sigma);
    gwy_synth_save_arg_double(settings, key, "tau", pargs->tau);
    g_string_free(key, TRUE);
}

static void
reset_star(gpointer p)
{
    PatSynthControlsStar *pcontrols = p;
    PatSynthArgsStar *pargs = pcontrols->args;

    *pargs = pat_synth_defaults_star;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->n_rays),
                             pargs->n_rays);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac),
                             pargs->top_frac);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->top_frac_noise),
                             pargs->top_frac_noise);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->edge_shift),
                             pargs->edge_shift);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->height), pargs->height);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->xcenter),
                             pargs->xcenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->ycenter),
                             pargs->ycenter);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->angle), pargs->angle);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->sigma), pargs->sigma);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(pcontrols->tau), pargs->tau);
}

/***************************************************************************
 *
 * Common settings
 *
 ***************************************************************************/

static const gchar active_page_key[] = "/module/pat_synth/active_page";
static const gchar update_key[]      = "/module/pat_synth/update";
static const gchar randomize_key[]   = "/module/pat_synth/randomize";
static const gchar seed_key[]        = "/module/pat_synth/seed";
static const gchar type_key[]        = "/module/pat_synth/type";

static void
pat_synth_sanitize_args(PatSynthArgs *args)
{
    args->active_page = CLAMP(args->active_page,
                              PAGE_DIMENSIONS, PAGE_NPAGES-1);
    args->update = !!args->update;
    args->seed = MAX(0, args->seed);
    args->randomize = !!args->randomize;
    args->type = MIN(args->type, PAT_SYNTH_NTYPES-1);
}

static void
pat_synth_load_args(GwyContainer *container,
                    PatSynthArgs *args,
                    GwyDimensionArgs *dimsargs)
{
    *args = pat_synth_defaults;

    gwy_container_gis_int32_by_name(container, active_page_key,
                                    &args->active_page);
    gwy_container_gis_boolean_by_name(container, update_key, &args->update);
    gwy_container_gis_int32_by_name(container, seed_key, &args->seed);
    gwy_container_gis_boolean_by_name(container, randomize_key,
                                      &args->randomize);
    gwy_container_gis_enum_by_name(container, type_key, &args->type);
    pat_synth_sanitize_args(args);

    gwy_clear(dimsargs, 1);
    gwy_dimensions_copy_args(&dims_defaults, dimsargs);
    gwy_dimensions_load_args(dimsargs, container, prefix);
}

static void
pat_synth_save_args(GwyContainer *container,
                    const PatSynthArgs *args,
                    const GwyDimensionArgs *dimsargs)
{
    gwy_container_set_int32_by_name(container, active_page_key,
                                    args->active_page);
    gwy_container_set_boolean_by_name(container, update_key, args->update);
    gwy_container_set_int32_by_name(container, seed_key, args->seed);
    gwy_container_set_boolean_by_name(container, randomize_key,
                                      args->randomize);
    gwy_container_set_enum_by_name(container, type_key, args->type);

    gwy_dimensions_save_args(dimsargs, container, prefix);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
